/*
 *
 * main/AScreen.H --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef AScreen_H
#define AScreen_H

#include "config.h"

#include <nucleo/gl/window/glWindow.H>
#include <nucleo/gl/scenegraph/sgNode.H>
#include <nucleo/gl/scenegraph/sgViewpoint.H>
#include <nucleo/image/sink/ImageSink.H>

#include "wstyle/AbstractWindowRenderer.H"
#include "fvwmmodule/FvwmModule.H"

#include "LayerManager.H"
#include "AUtils.H"

#define DRAG_ROTATE_NO 0
#define DRAG_ROTATE_Z  1
#define DRAG_ROTATE_X  2
#define DRAG_ROTATE_Y  3
#define DRAG_ROTATE_XY 4

class XwncDesktop;

using namespace nucleo ;

class AScreen {

 private:
	DefaultSettings *_settings;
	ImageSink *_capture;
	Image _capture_image;
	bool _useTransparency;
	bool _orthoProj;
	bool _shapeWindow;
	bool _focusHilite;
	PerfMeter *_perfMeter;
	ACursor *_cursor;
	GLint _winwidth;
	GLint _winheight;
	unsigned char _button_mask;
	bool _motion;
	bool _moving;
	bool _rotate;
	int _rotateType;
	bool _folding;
	int _prevX;
	int _prevY;
	bool _breakDisplayLists;
	bool _saveFrame;
	bool _isFvwmModule;

	bool _inPseudoSelectionMode;

	bool _wmMove;
	bool _wmResize;
	float _resizeTransparency;
	int _opsDx;
	int _opsDy;
	XwncWindow *opWin;

	// internal move
	bool _inInternalMove;
	XwncWindow *_imWindow;
	int _imStartX, _imStartY;
	bool _waitForAConfigureWindow;
	bool _inClickMotion;
	XwncWindow *_clickMotionWindow;

	sgViewpoint *_viewpoint;

	AbstractWindowRenderer *_selection;
	GLuint _selectionBuffer[256] ;
	int _selectionBufferSize;
	bool _resetSelectionScheduled;

	bool _noStencil;

	FvwmModule *_fvwmModule;
	XwncDesktop *_wncDesktop;

	void _myPointerEvent(int x, int y);
	int _myPickClosest(int x, int y, GLuint *sbuffer, GLuint size);
	void _x11project(int x, int y, int *px, int *py);
	void _x11projectOverWindow(
		XwncWindow *win, int x, int y, int *px, int *py);
	void _interactiveMove(int ex, int ey);
	void _interactiveResize(int ex, int ey);
	void _exposeModeHandleEvents(void);
	void _internalMove(int x, int y);
	void _stopInternalMove(bool cancel = false);
	void  _stopClickMotion(void);

public:
	glWindow *ametista;
	LayerManager *root;
  
	AScreen(XwncDesktop *wncDesktop, bool ortho, bool shaped, bool focusHi,
		bool noStencil);

	void setTitle(char *title);
	void setGeometry(unsigned int width, unsigned int height);
	void makeCurrent(void);
	void showCursor(void);
	void initCapture(char *captureSink);
	void saveFrame(void);
	void showFPS(void);
	bool hasStencilBuffer(void);
	bool doUsePerVncWinTexture(void);
	void setUsePerVncWinTexture(bool v);
	void setTransparent(bool transparent);
	void toggleTransparent(void);
	void resetSelection(bool sendEvend = true);
	void resetSelection(int x, int y);
	void scheduleResetSelection(void);
	bool doPseudoSelection(void);
	bool  unprojectOverWindow(
		XwncWindow *win, int x, int y, GLdouble *ox,
		GLdouble *oy, GLdouble *oz);
	bool projectOverWindow(
		XwncWindow *win, int x, int y, GLdouble *px, GLdouble *py,
		GLdouble *pz);
	void _myPointerEventOverWindow(XwncWindow *win, int x, int y);
	void setXvncDesktop(sgNode *d);
	void setFvwmModule(FvwmModule *m);
	void wmMove(XwncWindow *w, bool start);
	void wmResize(XwncWindow *w, bool start);
	void setResizeTransparency(float alpha);
	bool waitingForAConfigureWindow(void);
	bool inInteractiveMove(void);
	bool inInteractiveResize(void);
	void  clickMotion(XwncWindow *win);
	void fold(XwncWindow *win, int x, int y);
	void dragRotate(XwncWindow *win, int x, int y, int type);
	void dragRotate(XwncWindow *win, int x, int y);
	void dragRotate(XwncWindow *win, int type);
	void startInternalMove(
		XwncWindow *win, int x = -1, int y = -1, float alpha = 0.7);
	int handleEvents(void);
	void draw(bool force);
	void deleteCapture(void);
};

#endif
