/*
 *
 * main/AScreen.cxx --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include <math.h>
#include <stdio.h>

#include <nucleo/utils/AppUtils.H>
#include <nucleo/gl/window/glWindow.H>
#include <nucleo/gl/glUtils.H>
#include <nucleo/image/sink/ImageSink.H>
#include <nucleo/utils/FileUtils.H>
#include <nucleo/image/encoding/Conversion.H>

#include "wncdesktop/XwncDesktop.H"

#include "AUtils.H"
#include "AScreen.H"
using namespace nucleo ;

void
AScreen::setTitle(char *title)
{
	ametista->setTitle(title);
}

void
AScreen::setGeometry(unsigned int width, unsigned int height)
{
	if (width == ametista->getScreenWidth() &&
	    height == ametista->getScreenHeight())
	{
	  ametista->setFullScreen(true);
	}
	else
	{
		ametista->setGeometry(width, height, 0, 0);
	}
	ametista->setMinMaxSize(width, height, width, height);
	_winwidth = width ;
	_winheight = height ;
}

void AScreen::makeCurrent(void)
{
	ametista->makeCurrent();
}

void AScreen::showCursor()
{
	ametista->setCursorVisible(false);
	_cursor = new ACursor("cursor", 5);
	_viewpoint->addDependency(_cursor);
	_wncDesktop->setSoftwareCursor(_cursor);
}

void AScreen::initCapture(char *captureSink)
{
  //_capture = createImageSink(captureSink) ;
	//ametista->addNode(_capture) ;
	//_capture->start() ;
}

void AScreen::showFPS(void)
{
	_perfMeter = new PerfMeter;
	_viewpoint->addDependency(_perfMeter);  
}

void AScreen::saveFrame(void)
{
	_saveFrame = true;
}

void AScreen::setTransparent(bool transparent) {
	_useTransparency =  transparent;
	if (transparent)
	{
		glEnable(GL_BLEND);
	}
	else
	{
		glDisable(GL_BLEND) ;
	}
	_viewpoint->postRedisplay() ;
	_breakDisplayLists = true ;
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void AScreen::toggleTransparent(void)
{
	_useTransparency = !_useTransparency ;
	setTransparent(_useTransparency);
}

void AScreen::setFvwmModule(FvwmModule *m)
{
	_fvwmModule = m;
}

void AScreen::deleteCapture(void)
{
	delete _capture;
}

bool AScreen::hasStencilBuffer(void)
{
	return !_noStencil;
}

// ---------------------------------------------------------------------------
// Selection and mouse events

#define DEBUG_SELECTION 0

bool AScreen::doPseudoSelection(void)
{
	return _inPseudoSelectionMode;
}

void AScreen::_myPointerEvent(int x, int y)
{
	GLdouble ox, oy, oz;

	if (_selection && _selection->getWindow() &&
	    unprojectOverWindow(
		    _selection->getWindow(), x, y, &ox, &oy, &oz))
	{
		// to make tooltips output only
		_selection->pointerEvent((int)ox,(int)oy, _button_mask);
	}
	else if (_selection)
	{
		unproject(
			x, y, _viewpoint, _selectionBuffer,
			_selectionBufferSize, &ox, &oy, &oz) ;
		_selection->pointerEvent((int)ox,(int)oy, _button_mask);
	}
	else
	{
		_wncDesktop->rootPointerEvent(x, y, _button_mask);
	}
}

// we need our own PickClosest for shaped window
int AScreen::_myPickClosest(int x, int y, GLuint *sbuffer, GLuint size)
{
	GLuint *tmpbuffer = new GLuint [sgViewpoint::glPickingBufferSize] ;

	int hits = _viewpoint->pickAll(
		x,y,tmpbuffer,sgViewpoint::glPickingBufferSize) ;
	if (!hits) {
		delete [] tmpbuffer ;
		return 0 ;
	}

	 GLuint *selection=0, selectionSize=0 ;
	 GLdouble minimum=0 ;
#if 0
	 std::cerr << "hits : "<< hits << std::endl;
#endif
	 for (GLuint *hit=tmpbuffer, i=0; (GLint)i<hits; i++)
	 {
		 GLuint numNames = *hit ;
		 GLdouble minWinZ = *(hit+1) / 4294967295.0 ;
		 // GLdouble maxWinZ = *(hit+2) / 4294967295.0 ;
		 GLuint *names = hit+3 ;
		 if (i==0 || minWinZ<=minimum || selectionSize == 0)
		 {
			 // ckeck
		   
			 GLdouble ox, oy, oz;
			 AbstractWindowRenderer *sel =
				 (AbstractWindowRenderer *)names[numNames-1];
#if 0
			 std::cerr << "    check : "
				   << sel->getWindow()->getTitle()
				   << " " << minWinZ 
				   << std::endl;
#endif
			 for (GLuint n=0; n<size && n<numNames; n++)
				 sbuffer[n] = names[n] ;
			 unproject(
				 x, y,minWinZ, _viewpoint, sbuffer,
				 numNames, &ox, &oy, &oz);
			 if (sel->isInsideWindow((int)ox,(int)oy)) 
			 {
				 selection = names ;
				 selectionSize = numNames ;
				 minimum = minWinZ ;
			 }
		 }
		 hit += 3+numNames ;
	 }

	 for (GLuint n=0; n<size && n<selectionSize; n++)
		 sbuffer[n] = selection[n] ;
	 delete [] tmpbuffer ;

	 return selectionSize ;    
}
   
void AScreen::resetSelection(bool sendEvent)
{
	AbstractWindowRenderer *oldSelection = _selection;
	_selectionBufferSize = 0;

#if  DEBUG_SELECTION && 0
	std::cerr << "Check Selection" << std::endl;
#endif
	_selectionBufferSize =
		_myPickClosest(_prevX,_prevY,_selectionBuffer, 256);

	if (_selectionBufferSize)
	{
		if (_focusHilite && _selection)
		{
			_selection->dim() ;
		}
		_selection = (AbstractWindowRenderer *)
			_selectionBuffer[_selectionBufferSize-1] ;
		if (_focusHilite && _selection)
		{
			_selection->undim();
		}
		if (oldSelection != _selection)
		{
#if DEBUG_SELECTION
			std::cerr << "New selection: "
				  << _selection->getName()
				  << " t: "
				  << _selection->getWindow()->getTitle()
				  << std::endl ;
#endif
		}
	}
	else
	{
		if (_selection)
		{
			if (_focusHilite)
			{
				_selection->dim() ;
			}
			_selection = 0 ;
#if DEBUG_SELECTION
			std::cerr << "No selection" << std::endl ;
#endif
		}
	}
	if (sendEvent)
	{
		_myPointerEvent(_prevX, _prevY);
	}
}

void AScreen::resetSelection(int x, int y)
{
	_prevX = x;
	_prevY = y;
	resetSelection();
}


void AScreen:: scheduleResetSelection(void)
{
	_resetSelectionScheduled = true;
}

// --------------------------------------------------------------------------
// Send event and unproject over a window which is not necessarily the selection
// (e.g., a window under the selection).
// At present time only used with the root window (not sure that this can
// work with a window with some XY rotation).

void AScreen::_myPointerEventOverWindow(XwncWindow *win, int x, int y)
{
	GLdouble ox, oy, oz;

	if (unprojectOverWindow(win, x, y, &ox, &oy, &oz))
	{
		win->getRenderer()->pointerEvent((int)ox,(int)oy, _button_mask);
#if 0
		std::cerr << "mpeow: " << x << " " << y << " / "
			  << ox << " " << oy << std::endl;
#endif
	}
	else if (unprojectOverWindow(root->getRootWindow(), x, y, &ox, &oy, &oz))
	{
		win->getRenderer()->pointerEvent((int)ox,(int)oy, _button_mask);
		//std::cerr << "mpeow: fail on win" << std::endl;
	}
	else
	{
		_wncDesktop->rootPointerEvent(x, y, _button_mask);
		//std::cerr << "mpeow: fail" << std::endl;
	}
}

bool AScreen::unprojectOverWindow(
	XwncWindow *win, int x, int y, GLdouble *ox, GLdouble *oy,
	GLdouble *oz)
{
	// should found the z. a simpler method? not sure that this works
	// with window with some XY rotation.

	_inPseudoSelectionMode = true;
	float z = 0;
	GLuint *tmpbuffer = new GLuint [sgViewpoint::glPickingBufferSize] ;
	int hits = _viewpoint->pickAll(
		x,y,tmpbuffer,sgViewpoint::glPickingBufferSize) ;
	_inPseudoSelectionMode = false;

	if (!hits) {
		delete [] tmpbuffer ;
		//std::cerr << "uow: fail" << std::endl;
		return false;
	}

	bool found = false;
	for (GLuint *hit=tmpbuffer, i=0; (GLint)i<hits; i++)
	{
		 GLuint numNames = *hit ;
		 GLuint *names = hit+3 ;
		 XwncWindow *w = (XwncWindow *)names[numNames-2];
		 if (w == win)
		 {
			 z = *(hit+1) / 4294967295.0 ;
			 found = true;
			 break;
		 }
		 hit += 3+numNames ;
	}
	delete [] tmpbuffer ;

	if (!found)
	{
		//std::cerr << "uow: not found" << std::endl;
		return false;
	}

	// FIXME, this should not be hard coded
	GLuint selBuffer[3] ;
	GLuint selSize=3;

	selBuffer[2] = (GLuint )(win->getRenderer());
	selBuffer[1] = (GLuint )(win);
	selBuffer[0] = (GLuint )(root);

	unproject(x, y, z, _viewpoint, selBuffer, selSize, ox, oy, oz) ;
#if 0
	std::cerr << "uow: " << x << " " << y << " " << z << " / "
			  << *ox << " " << *oy << " " << *oz << std::endl;
#endif
	return true;
}

bool AScreen::projectOverWindow(
	XwncWindow *win, int x, int y, GLdouble *px, GLdouble *py, GLdouble *pz)
{
	float sw,sh;
	float sx, sy;
	win->getSize(&sw,&sh);
	win->getRealPosition(&sx,&sy);
	// object coordinates
	float ox, oy;
	ox = (float)x - sx - sw/2;
	oy = - (float)y + sy + sh/2;

	GLuint selBuffer[3] ;
	GLuint selSize=3;
	selBuffer[2] = (GLuint )(win->getRenderer());
	selBuffer[1] = (GLuint )(win);
	selBuffer[0] = (GLuint )(root);

	_viewpoint->project(
			ox, oy, 0.0, selBuffer, selSize, px, py, pz);
	return true;
}

// --------------------------------------------------------------------------
// Window manager operations

void AScreen::wmMove(XwncWindow *w, bool start)
{
	if (start)
	{
		w->setTmpAlpha(0.5);
		_wmMove = true;
		opWin = w;
#if 0
		GLdouble ox, oy, oz;
		unprojectOverWindow(opWin, _prevX, _prevY, &ox, &oy, &oz);
		float rx,ry,rh,rw;
		opWin->getRealPosition(&rx, &ry);
		opWin->getSize(&rw, &rh);

		float realx = ox + rw/2.0 + rx;
		float realy = -oy + rh/2.0 + ry;
		GLdouble  px, py, pz;
		unprojectOverWindow(
			root->getRootWindow(), _prevX, _prevY, &px, &py, &pz);
		float rootx = (int)(px + _winwidth/2.0);
		float rooty = (int)(-py + _winheight/2.0);
		_opsDx = (int)(realx - rootx);
		_opsDy = (int)(realy - rooty);
#endif
	}
	else
	{
		w->resetAlpha();
		_wmMove = false;
		opWin = 0;
		resetSelection();
	}
}

void AScreen::_interactiveMove(int ex, int ey)
{
	_myPointerEventOverWindow(opWin, ex,  ey);
	_prevX = ex;
	_prevY = ey;
}

bool AScreen::inInteractiveMove(void)
{
	return _wmMove;
}

void AScreen::setResizeTransparency(float alpha)
{
	_resizeTransparency = alpha;
}

void AScreen::wmResize(XwncWindow *w, bool start)
{
	if (start)
	{
		w->setTmpAlpha(_resizeTransparency);
		_wmResize = true;
		opWin = w;
#if 0
		GLdouble ox, oy, oz;
		unprojectOverWindow(
			opWin, _prevX, _prevY, &ox, &oy, &oz);
		float rx,ry,rh,rw;
		opWin->getRealPosition(&rx, &ry);
		opWin->getSize(&rw, &rh);
		float realx = ox + rw/2.0 + rx;
		float realy = -oy + rh/2.0 + ry;
		GLdouble  px, py, pz;
		unprojectOverWindow(
			root->getRootWindow(), _prevX, _prevY, &px, &py, &pz);
		float rootx = px + _winwidth/2.0;
		float rooty = -py + _winheight/2.0;
		_opsDx = (int)(realx - rootx);
		_opsDy = (int)(realy - rooty);
#endif
	}
	else
	{
		w->resetAlpha();
		_wmResize = false;
		opWin = 0;
		resetSelection();
	}
}

void AScreen::_interactiveResize(int ex, int ey)
{
	_myPointerEventOverWindow(opWin, ex,  ey);
	_prevX = ex;
	_prevY = ey;
}

bool AScreen::inInteractiveResize(void)
{
	return _wmResize;
}

void  AScreen::clickMotion(XwncWindow *win)
{
	if (_button_mask == 0)
	{
		return;
	}
	if (!_selection || win != _selection->getWindow())
	{
		return;
	}
	_clickMotionWindow = win;
	_inClickMotion = true;
	// FIXME: should be configurable
	root->setTmpAlphaAbove(win, true, 0.5);
}

void  AScreen::_stopClickMotion(void)
{
	root->setTmpAlphaAbove(_clickMotionWindow, false);
	_inClickMotion = false;
}

// ----------------------------------------------------------------------------
// ametista operation

void AScreen::_x11project(int x, int y, int *px, int *py)
{
	if (_selection)
	{
		GLdouble fx, fy, fz;
		float sw,sh;
		float sx, sy;
		_selection->getWindow()->getSize(&sw,&sh);
		_selection->getWindow()->getRealPosition(&sx,&sy);
		// object coordinates
		float ox, oy;
		ox = (float)x - sx - sw/2;
		oy = - (float)y + sy + sh/2;
		_viewpoint->project(
			ox, oy, 0.0, _selectionBuffer, _selectionBufferSize,
			&fx, &fy, &fz);
		// window coor to X11 coor
		*px = (int)(fx);
		*py = (int)(_winheight - fy);
	}
	else
	{
		*px = x;
		*py = y;
	}
}

void AScreen::_x11projectOverWindow(
	XwncWindow *win, int x, int y, int *px, int *py)
{
	GLdouble fx,fy,fz;
	projectOverWindow(win, x, y, &fx, &fy, &fz);
	// window coor to X11 coor
	*px = (int)(fx);
	*py = (int)(_winheight - fy);
}

void AScreen::fold(XwncWindow *win, int x, int y)
{
	GLdouble ox, oy, oz ;
	int xx, yy;
	opWin = win;
	_x11projectOverWindow(win, x, y, &xx, &yy);
	unprojectOverWindow(
		opWin, xx, yy, &ox, &oy, &oz) ;
	opWin->getRenderer()->fold((int)ox,(int)oy, true);
	_folding = true;
}

void AScreen::dragRotate(XwncWindow *win, int x, int y, int type)
{
	//if (!_selection || win != _selection->getWindow()) return;
	
	if (!_rotate)
	{
		_rotate = true;
		_rotateType = type;
		opWin = win;
	}
	else
	{
		float rx = x - _prevX;
		float ry = y - _prevY;

		switch(_rotateType)
		{
		case DRAG_ROTATE_Z:
			if ((rx > 0 && ry < 0) || (rx < 0 && ry < 0))
			{
				//root->rotateWindow(
				//	opWin, -rx/2, 0, 0, -ry/2);
				opWin->getRenderer()->rotate_rel(
					-rx/2, 0, 0, -ry/2);
			}
			else if ((rx < 0 && ry > 0) || (rx > 0 && ry > 0))
			{
				//root->rotateWindow(
				//	opWin, -rx/2, 0, 0, ry/2);
				opWin->getRenderer()->rotate_rel(
					-rx/2, 0, 0, ry/2);
			}
			break;
		case DRAG_ROTATE_X:
			if (ry < 0)
			{
				root->rotateWindow(
					opWin,ry/2, -ry/2, 0, 0);
				//opWin->getRenderer()->rotate_rel(
				//	ry/2, -ry/2, 0, 0);
			}
			else if (ry > 0)
			{
				root->rotateWindow(
					opWin, ry/2, ry/2, 0, 0);
				//opWin->getRenderer()->rotate_rel(
				//	ry/2, ry/2, 0, 0);
			}
			break;
		case DRAG_ROTATE_Y:
			if (rx > 0)
			{
				root->rotateWindow(opWin, rx/2, 0, rx/2, 0);
				//opWin->getRenderer()->rotate_rel(
				//		rx/2, 0, rx/2, 0);
			}
			else if (rx < 0)
			{
				root->rotateWindow(opWin,rx/2, 0, -rx/2, 0);
				//opWin->getRenderer()->rotate_rel(
				//	rx/2, 0, -rx/2, 0);
			}
			break;
		default: 
		case DRAG_ROTATE_XY:
			if (ry < 0)
			{
				root->rotateWindow(opWin,ry/2, -ry/2, 0, 0);
				//opWin->getRenderer()->rotate_rel(
				//ry/2, -ry/2, 0, 0);
			}
			else if (ry > 0)
			{
				root->rotateWindow(opWin,ry/2, ry/2, 0, 0);
				//opWin->getRenderer()->rotate_rel(
				//ry/2, ry/2, 0, 0);
			}
			if (rx > 0)
			{
				root->rotateWindow(opWin, rx/2, 0, rx/2, 0);
				//opWin->getRenderer()->rotate_rel(
				//	rx/2, 0, rx/2, 0);
			}
			else if (rx < 0)
			{
				root->rotateWindow(opWin,rx/2, 0, -rx/2, 0);
				//opWin->getRenderer()->rotate_rel(
				//	rx/2, 0, -rx/2, 0);
			}
			break;
		}
	}
}

void AScreen::dragRotate(XwncWindow *win, int type)
{
	_rotate = false;
	opWin = 0;
	dragRotate(win, 0,0,type);
}

void AScreen::dragRotate(XwncWindow *win, int x, int y)
{
	dragRotate(win, x,y,_rotateType);
}

bool AScreen::waitingForAConfigureWindow(void)
{
	bool rc = false;

	if (_waitForAConfigureWindow)
	{
		rc = true;
		_waitForAConfigureWindow = false;
	}
	return rc;
}

void AScreen::_internalMove(int x, int y)
{
	int oldX, oldY;
	_imWindow->getRealPosition(&oldX, &oldY);

	if (x != _prevX || y != _prevY)
	{
		_imWindow->translate_rel(-_prevX + x, _prevY - y, 0);
		_imWindow->setRealPosition(oldX -_prevX + x, oldY - _prevY + y);
	}
}

void AScreen::startInternalMove(XwncWindow *win, int x, int y, float alpha)
{
	if (_inInternalMove)
	{
		return;
	}

#if 0
	std::cerr << "Start Internal Move" << std::endl;
#endif
	win->setTmpAlpha(alpha);
	win->getRealPosition(&_imStartX, &_imStartY);
	_inInternalMove = true;
	_imWindow = win;

	int xx,yy;

	_x11projectOverWindow(win, x, y, &xx, &yy);
	//_x11project(x, y, &xx, &yy);

	int savePrevX = _prevX;
	int savePrevY = _prevY;
	_prevX = xx;
	_prevY = yy;
	_internalMove(savePrevX, savePrevY);
	_prevX = savePrevX;
	_prevY = savePrevY;
}

void AScreen::_stopInternalMove(bool cancel)
{
	int x,y;
	char cmd[256];
	
#if 0
	std::cerr << "Stop Internal Move" << std::endl;
#endif

	_imWindow->getRealPosition(&x, &y);
	if (cancel)
	{
		_imWindow->translate_rel(_imStartX - x, - _imStartY + y, 0);
		_imWindow->setRealPosition(_imStartX, _imStartY);
		_inInternalMove = false;
		resetSelection();
	}
	else
	{
		sprintf(cmd, "Move w%ip w%ip", x - _imStartX, y - _imStartY);
		_fvwmModule->sendText(cmd, _imWindow);
		_waitForAConfigureWindow = true;
		_inInternalMove = false;
		int savePrevX = _prevX;
		int savePrevY = _prevY;
		_prevX = _prevX-1;
		_prevY = _prevY+1;
		_myPointerEventOverWindow(root->getRootWindow(), _prevX, _prevY);
		_prevX = savePrevX;
		_prevY = savePrevY;
	}
	_imWindow->resetAlpha();
	_inInternalMove = false;
	_imWindow = 0;
}

// -----------------------------------------------------------------------------
// events handling

void AScreen::_exposeModeHandleEvents(void)
{
	static bool first = true;
	bool recover = false;
	glWindow::event e ;
	
	if (first)
	{
		resetSelection();
		first = false;
	}
	while (ametista->getNextEvent(&e) && !recover)
	{
		switch (e.type)
		{
		case glWindow::event::configure:
			break;
		case glWindow::event::expose:
			_viewpoint->postRedisplay();
			break ;
		case glWindow::event::buttonPress:
			_button_mask = _button_mask | (1<<(e.button-1)) ;
			_prevX = e.x ;
			_prevY = e.y ;
			if (_perfMeter) _perfMeter->restart();
			break;
		case glWindow::event::pointerMotion:
			if (_cursor)
			{
				_cursor->updatePosition(
					e.x-_winwidth/2.0, _winheight/2.0-e.y) ;
			}
			_prevX = e.x ;
			_prevY = e.y ;
			if (_button_mask == 0)
				resetSelection();
			break;
		case glWindow::event::buttonRelease:
			_button_mask = _button_mask & (~(1<<(e.button-1))) ;
			recover = true;
			break;
		case glWindow::event::keyPress:
			break;
		case glWindow::event::keyRelease:
			if (e.keysym == XK_Escape)
				recover = true;
			break;
		case glWindow::event::leave:
			break;
		default:
			break ;	
		}
	}

	if (recover)
	{
		root->exposeMode();
		draw(false);
		if (_selection && !_selection->getWindow()->isRootWindow())
		{
			_fvwmModule->sendText(
				"AFuncDeiconifyFocusAndRaise",
				_selection->getWindow());
		}
		resetSelection();
		_wncDesktop->releaseModifiers();
		first = true;
	}
}

int AScreen::handleEvents(void)
{
	static bool ready = false;
	int loop = 1;

	if (_perfMeter)
	{
	  _perfMeter->postRedisplay() ;
	}


	if (root->inExposeMode())
	{
		_exposeModeHandleEvents();
		return 1;
	}

	glWindow::event e ;
	while (ametista->getNextEvent(&e))
	{
		_motion = 0;
		switch (e.type)
		{
		case glWindow::event::configure:
		{
#if 0
			std::cerr << "Configure Window: " << _winwidth << " c: "
				  << e.width << " " << _winheight <<  " c: "
				  << e.height << " " << ready << std::endl;
#endif
			if (_winwidth != e.width || _winheight != e.height)
			{
				break;
			}
			if (!ready)
			{
				_wncDesktop->startRequest();
				ready = true;
			}
			double far;
			double near;
			double eyeDist;
			glViewport(0,0,e.width,e.height) ;
			if (_orthoProj)
			{
				GLfloat w2=e.width/2.0, h2=e.height/2.0 ;
				((sgOrthoViewpoint *)_viewpoint)->resize(
					-w2,w2,-h2,h2,-5000,5000) ;
				eyeDist = 0;
				near = -5000;
				far = 5000;
			}
			else
			{
				double fovy = 60.0 ;
				double a = (fovy*M_PI/180.0)/2.0 ;
				double h = (double)e.height / (2.0*sin(a)) ;
				double dist = sqrt(
					h*h - (double)(e.height*e.height)/4.0) ;
				double r = (GLfloat)e.height/(GLfloat)e.width;
				near = dist;
				// FIXME: should be configurable with an option
				far = dist*5;
				((sgPerspectiveViewpoint *)_viewpoint)->
					setNearLimit(near) ;
				((sgPerspectiveViewpoint *)_viewpoint)->
					setFarLimit(far) ;
				((sgPerspectiveViewpoint *)_viewpoint)->
					setEyePoint(0.0, 0.0, dist) ;
				((sgPerspectiveViewpoint *)_viewpoint)->
					resetTransformations() ;
				((sgPerspectiveViewpoint *)_viewpoint)->
					scale(r, 1.0, 1.0) ;
				eyeDist = dist;
			}
			root->setDesktopProperties(
				(float)_winwidth, (float)_winheight, eyeDist,
				near, far, _viewpoint);
			if (_perfMeter)
			{
				_perfMeter->resetTransformations() ;
				_perfMeter->translate(
					-_winwidth/2+5, -_winheight/2+5, 0) ;
			}
		}
		break ;
		case glWindow::event::expose:
			_viewpoint->postRedisplay();
			break ;
		case glWindow::event::buttonPress:
			_button_mask = _button_mask | (1<<(e.button-1)) ;
			if (e.x < 0) e.x = 0;
			if (e.x > _winwidth -1) e.x = _winwidth -1;
			if (e.y < 0) e.y = 0;
			if (e.y > _winheight -1) e.y = _winheight -1;
#if 0
			std::cerr << "buttonPress: " << e.button
				  << " (status=" << (int)_button_mask << ")"
				  << std::endl ;
#endif
			if (_wmMove)
			{
				_interactiveMove(e.x, e.y);
			}
			else if (_wmResize)
			{
				_interactiveResize(e.x, e.y);
			}
			else if (_inInternalMove)
			{
				_stopInternalMove();
			}
			else
			{
				_myPointerEvent(e.x,e.y);
			}
			_prevX = e.x ;
			_prevY = e.y ;
			if (_perfMeter) _perfMeter->restart() ;	
			break;
		case glWindow::event::pointerMotion:
		{
			float tx = 0;
			float ty = 0;
			if (e.x < 0) e.x = 0;
			if (e.x > _winwidth -1) e.x = _winwidth -1;
			if (e.y < 0) e.y = 0;
			if (e.y > _winheight -1) e.y = _winheight -1;
			if (e.x < 4) { tx = 10; }
			if (e.y < 4) { ty = -10; }
			if (e.x > _winwidth - 5) { tx = -10; }
			if (e.y > _winheight - 5) { ty = 10; }
			if (tx != 0 || ty != 0)
			{
				root->translate_rel(tx,ty,0);
			}
			_motion = 1;
			if (_cursor)
			{
				_cursor->updatePosition(
					e.x-_winwidth/2.0, _winheight/2.0-e.y) ;
			}
			if (_wmMove)
			{
				_motion = 0;
				_interactiveMove(e.x, e.y);
			}
			else if (_wmResize)
			{
				_motion = 0;
				_interactiveResize(e.x, e.y);
			}
			else if (_inInternalMove)
			{
				_motion = 0;
				_internalMove(e.x, e.y);
			}
			else if (_folding && opWin) {
				_motion = 0;
				GLdouble ox, oy, oz ;
				unprojectOverWindow(
					opWin,e.x, e.y, &ox, &oy, &oz) ;
				opWin->getRenderer()->fold((int)ox,(int)oy, true);
			} 
			else if (_rotate && opWin)
			{
				_motion = 0;
				dragRotate(_selection->getWindow(), e.x,e.y);
			}
			else if (0 && _button_mask &&_selection)
			{
				// At least one button is pressed
				// WE MUST REST THE SELECTION!
				_myPointerEvent(e.x,e.y);
			}
			else
			{
				resetSelection(e.x, e.y);
			}
			_prevX = e.x ;
			_prevY = e.y ;
		}
		break ;
		case glWindow::event::buttonRelease:
		{
			_button_mask = _button_mask & (~(1<<(e.button-1))) ;
			if (e.x < 0) e.x = 0;
			if (e.x > _winwidth -1) e.x = _winwidth -1;
			if (e.y < 0) e.y = 0;
			if (e.y > _winheight -1) e.y = _winheight -1;
#if 0
			std::cerr << "buttonRelease: " << e.button
				  << " (status=" << (int)_button_mask << ")"
				  << std::endl ;
#endif
			bool reselect = false;
			if (_inClickMotion)
			{
				_stopClickMotion();
			}
			if (_rotate)
			{
				reselect = true;
				_rotate = false;
				opWin = 0;
			}
			if (_folding)
			{
				if (opWin)
				{
					GLdouble ox, oy, oz ;
					unprojectOverWindow(
						opWin,e.x, e.y, &ox, &oy, &oz) ;
					opWin->getRenderer()->fold(
						(int)ox,(int)oy, false);
				}
				reselect = true;
				_folding = false;
				opWin = 0;
			}
			
			if (_wmMove)
			{
				_interactiveMove(e.x, e.y);
			}
			else if (_wmResize)
			{
				_interactiveResize(e.x, e.y);
			}
			else if (_inInternalMove)
			{
				_stopInternalMove();
			}
			else
			{
				if (reselect)
				{
					resetSelection(e.x, e.y);
				}
				else
				{
					_myPointerEvent(e.x,e.y);
				}
			}
			_prevX = e.x ;
			_prevY = e.y ;
		}
		break ;
		case glWindow::event::keyPress:
		{
			bool stopOp = false;
			bool cancelOp = false;

			if (_inInternalMove || _folding || _rotate)
			{
				switch(e.keysym)
				{
				case XK_Return:
				case XK_KP_Enter:
				case XK_space:
					stopOp = true;
					cancelOp = false;
					break;
				case XK_Escape:
					stopOp = true;
					cancelOp = true;
					break;
				default:
					break;
				}
				// FIXME: handle keyboard move with the usual
				// keys, see fvwm/libs/Target.c
			}
			if (stopOp)
			{
				if (_inInternalMove)
				{
					_stopInternalMove(cancelOp);
				}
				else if (_folding)
				{
					if (opWin)
					{
						GLdouble ox, oy, oz ;
						unprojectOverWindow(
							opWin,e.x, e.y,
							&ox, &oy, &oz) ;
						opWin->getRenderer()->fold(
							(int)ox,(int)oy, false);
					}
					_folding = false;
					opWin = 0;
				}
				_rotate = false;
				opWin = 0;
			}
			if (_selection)
			{
				_selection->keyEvent(e.keysym, true) ;
			}
			else
			{
				_wncDesktop->rootKeyEvent(e.keysym, true);
			}
			_wncDesktop->setModifiers(e, true);
		}
		break ;
		case glWindow::event::keyRelease:
			if (_selection)
			{
				_selection->keyEvent(e.keysym, false);
			}
			else
			{
				_wncDesktop->rootKeyEvent(e.keysym, false);
			}
			_wncDesktop->setModifiers(e, false);
			break;
		case glWindow::event::leave:
			_wncDesktop->releaseModifiers();
			break;
		default:
			break ;
		}

	}
	return loop;
}

// -----------------------------------------------------------------------------

void AScreen::draw(bool force)
{ 
	if (force || _viewpoint->graphChanged())
	{
		_viewpoint->displayGraph(
			_breakDisplayLists ? sgNode::NODL : sgNode::CREATE) ;
		if (_capture || _saveFrame)
		{
			glScreenCapture(&_capture_image, Image::RGB, true) ;
			if (_capture)
			{
				_capture->handle(&_capture_image) ;	
			}
			else
			{
				convertImage(&_capture_image, Image::JPEG) ;
				_capture_image.saveAs("capture.jpg") ;
				std::cerr << "Image saved" << std::endl ;
				_saveFrame = false;
			}
		}
		//std::cerr << "draw" << std::endl ;
		ametista->swapBuffers() ;

		if (_resetSelectionScheduled)
		{
			//std::cerr << "_resetSelectionScheduled" << std::endl ;
			_resetSelectionScheduled = false;
			resetSelection();
		}
	}

}

// -----------------------------------------------------------------------------

AScreen::AScreen(XwncDesktop *wncDesktop, bool ortho, bool shaped, bool focusHi,
		 bool noStencil)
{

	long options = glWindow::DOUBLE_BUFFER | glWindow::DEPTH;

	_noStencil = noStencil;
	if (!noStencil)
	{
		options = options | glWindow::STENCIL;
	}

	long eventmask = glWindow::event::configure
		| glWindow::event::expose
		| glWindow::event::pointerMotion
		| glWindow::event::buttonPress
		| glWindow::event::buttonRelease
		| glWindow::event::keyPress
		| glWindow::event::keyRelease
		| glWindow::event::leave;

	ametista = glWindow::create(options, eventmask) ;

	_shapeWindow = shaped;
	_focusHilite = focusHi;
	_wncDesktop = wncDesktop;

	_orthoProj = ortho;
	if (_orthoProj)
	{
		_viewpoint = new sgOrthoViewpoint("orthoproj");
	}
	else
	{
		_viewpoint = new sgPerspectiveViewpoint(
			"persproj",
			0.0, 0.0, 1.25,   // ex, ey, ez
			0.001, 1500.0) ;  // near, far
	}
  
	_settings = new DefaultSettings() ;
	_viewpoint->addDependency(_settings) ;
	glClearDepth(100.2) ;

	root = new LayerManager("layermanager");
	_viewpoint->addDependency(root) ;

	_selection = 0;
	_selectionBufferSize = 0;
	_resetSelectionScheduled = false;

	_useTransparency = true;
	glEnable(GL_BLEND) ;
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	_capture = 0;
	_perfMeter = 0;
	_cursor = 0;
	_winwidth = 0;
	_winheight = 0;
	_button_mask = 0;
	_motion = 0;
	_wmMove = false;
	_wmResize = false;
	_resizeTransparency = 0.7;
	_folding = false;
	_rotate = false;
	_inInternalMove = false;
	_waitForAConfigureWindow = false;
	_inClickMotion = false;
	_prevX=0;
	_prevY=0;
	_breakDisplayLists = true;
	_saveFrame = false;
	_inPseudoSelectionMode = false;
	opWin = 0;
	_imWindow = 0;
}
