#ifndef AUtils_H
#define AUtils_H

/*
 *
 * AUtils.H --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/core/TimeKeeper.H>
#include <nucleo/gl/scenegraph/sgNode.H>
#include <nucleo/gl/scenegraph/sgViewpoint.H>

#ifdef HAVE_GLFT
#include <glft/FontManager.hpp>
#endif

#include <X11/Xlib.h>

#include "wncdesktop/XwncWindow.H"
#include "wncdesktop/wncSource.H"

using namespace nucleo ;

// ---------------------------------------------------------------------------

class DefaultSettings : public sgNode {
protected:
  void display(dlPolicy policy) ;
public:
  DefaultSettings(void) ;
} ;


// ---------------------------------------------------------------------------

class ACursor : public sgNode {
private:
  GLfloat _x, _y, _size;
  GLfloat _xhot, _yhot, _width, _height;
  rfbXCursorColors _colors;
  unsigned char *_Xbuf;
  CARD32 *_ARGBbuf;
protected:
  void display(dlPolicy policy) ;
public:
  ACursor(std::string name, GLfloat size) ;
  void updatePosition(GLfloat x, GLfloat y) ;
  void handleXCursor(
	  int xhot, int yhot, int width, int height, char *buf,
	  rfbXCursorColors colors);
  void handleARGBCursor(
	  CARD32 xhot, CARD32 yhot, CARD32 width, CARD32 height,
	  XcursorPixel *buf);
} ;

// ---------------------------------------------------------------------------

class PerfMeter : public sgNode {
private:
  Chronometer _chrono ;
protected:
  void display(dlPolicy policy) ;
public:
  PerfMeter(void) : sgNode("PerfMeter",0) {
  }
  void restart(void) {
    _chrono.start() ;
  }
} ;

// ---------------------------------------------------------------------------

void
unproject(int x, int y,
	  sgViewpoint *viewpoint, GLuint *selectionBuffer, int selectionBufferSize,
	  GLdouble *ox, GLdouble *oy, GLdouble *oz) ;

void
unproject(int x, int y, float z, sgViewpoint *viewpoint, GLuint *selectionBuffer,
	  int selectionBufferSize,
	  GLdouble *ox, GLdouble *oy, GLdouble *oz);
// ---------------------------------------------------------------------------

void
ASetShape(Display *dpy, Window w,
	  sgViewpoint *viewpoint, GLuint *selectionBuffer, 
	  int width, int height);

#endif
