/*
 *
 * main/AUtils.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

// FIXME
#include <sstream>
#include <string>

#define HAVE_SHAPE 0

#if HAVE_SHAPE
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#endif /* HAVE_SHAPE */

#include <X11/Xlib.h>

#include "wncdesktop/XwncWindow.H"
#include "AUtils.H"

#define DEBUG_LEVEL 0

// --------------------------------------------------------------------

void
DefaultSettings::display(dlPolicy policy) {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT) ;
}

DefaultSettings::DefaultSettings(void) : sgNode("DefaultSettings") {
  glEnable(GL_CULL_FACE) ;
  glEnable(GL_DEPTH_TEST) ;

  // glDisable(GL_LIGHTING) ;
  // glShadeModel(GL_FLAT) ;

#if 0
  glEnable(GL_POINT_SMOOTH) ;
  glHint(GL_POINT_SMOOTH_HINT,GL_NICEST) ;
#endif

#if 0
  glEnable(GL_LINE_SMOOTH) ;
  glHint(GL_LINE_SMOOTH_HINT,GL_NICEST) ;
#endif

#if 0
  glEnable(GL_POLYGON_SMOOTH) ;
  glHint(GL_POLYGON_SMOOTH_HINT,GL_NICEST) ;
#endif

  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST) ;

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  //glEnable(GL_DITHER);
  //glDisable(GL_DITHER);
}

// --------------------------------------------------------------------

ACursor::ACursor(std::string name, GLfloat size) : sgNode(name) {
  _x = _y = 0.0 ;
  _size = size ;
  _ARGBbuf = 0;
  _Xbuf = 0;
  _width = _height = 0;
  _xhot = _yhot = 0;
}

void
ACursor::display(dlPolicy policy) {
	
	glDisable(GL_DEPTH_TEST) ;

	glTranslatef(_x,_y,-0.01);
	glColor4f(1,1,1,1);
	glBegin(GL_POINTS);
  
	if (_ARGBbuf)
	{
		unsigned char *data = (unsigned char *)_ARGBbuf;
		int k = 0;
		for(GLfloat i = 0; i < _height; i++)
		{
			for(GLfloat j = 0; j < _width; j++, k = k + 4)
			{
				// in fact we have "premultiplied" BGRA
				GLfloat f = (GLfloat)data[k+3]/255.0;
				f = (f==0)? 0:1/f;
				glColor4f(
					((GLfloat)data[k+2]*f/255.0), 
					((GLfloat)data[k+1]*f/255.0),
					((GLfloat)data[k]*f/255.0),
					(GLfloat)data[k+3]/255.0);
				GLfloat u = j - _xhot;
				GLfloat v = -i + _yhot;
				glVertex2f(u,v);
			}
		}
	}
	else if (_Xbuf)
	{
		int bytesPerRow = ((int)_width + 7) / 8;
		int bytesData = bytesPerRow * (int)_height;
	  
		int k = 0, l = 0, m = 0;
		for(GLfloat i = 0; i < _height; i++, k = k + bytesPerRow)
		{
			l = 0;
			m = k;
			for(GLfloat j = 0; j < _width; j++)
			{
				glColor4ub(
					(_Xbuf[m] & (1 << l))? 
					_colors.foreRed: _colors.backRed,
					(_Xbuf[m] & (1 << l))?
					_colors.foreGreen: _colors.backGreen,
					(_Xbuf[m] & (1 << l))?
					_colors.foreBlue: _colors.backBlue,
					(_Xbuf[bytesData+m] & (1 << l))? 255:0);
				GLfloat u = j - _xhot;
				GLfloat v = -i + _yhot;
				glVertex2f(u,v);
				l++;
				if (l>7)
				{
					l = 0;
					m++;
				}
			}
		}
	}
	else
	{
		// draw nothing
	}
	glEnd();
	glEnable(GL_DEPTH_TEST) ;
}

void
ACursor::updatePosition(GLfloat x, GLfloat y)
{
	_x = x ;
	_y = y ;
	
	postRedisplay() ;
}

void ACursor::handleXCursor(
	int xhot, int yhot, int width, int height, char *buf,
	rfbXCursorColors colors)
{
	if (_ARGBbuf)
	{
		free(_ARGBbuf);
		_ARGBbuf = 0;
	}
	if (_Xbuf)
	{
		free(_Xbuf);
		_Xbuf = 0;
	}
	_xhot = xhot;
	_yhot = yhot;
	_width = width;
	_height = height;
	_colors = colors;
	_Xbuf = (unsigned char *)buf;
	postRedisplay();
}

void ACursor::handleARGBCursor(
	CARD32 xhot, CARD32 yhot, CARD32 width, CARD32 height, XcursorPixel *buf)
{
	if (_ARGBbuf)
	{
		free(_ARGBbuf);
		_ARGBbuf = 0;
	}
	if (_Xbuf)
	{
		free(_Xbuf);
		_Xbuf = 0;
	}
	_xhot = xhot;
	_yhot = yhot;
	_width = width;
	_height = height;
	_ARGBbuf = (CARD32 *)buf;
	postRedisplay() ;
}

// --------------------------------------------------------------------

void
PerfMeter::display(dlPolicy policy) {
  glDisable(GL_DEPTH_TEST) ;
  glColor4f(0,0,0,1) ;

  _chrono.tick() ;
  std::stringstream tmp ;
  tmp << _chrono.average() << " fps" ;
  std::string label = tmp.str();

  if (_chrono.read() > 5000.0) {
    std::cerr << label << " (last " << _chrono.read()/1000.0
	      << " sec)" << std::endl;
    restart();
  }

}

// --------------------------------------------------------------------

void
unproject(int x, int y, sgViewpoint *viewpoint, GLuint *selectionBuffer,
	  int selectionBufferSize,
	  GLdouble *ox, GLdouble *oy, GLdouble *oz) {
  static int first_time = 1;
  static int hack = 1;
  GLfloat z = -10.0 ;

  GLint viewport[4] ;
  glGetIntegerv(GL_VIEWPORT, viewport) ;

  viewpoint->applyTransforms() ;

  GLdouble projmatrix[16] ;
  glGetDoublev(GL_PROJECTION_MATRIX, projmatrix) ;

  for (int i=0; i<selectionBufferSize; ++i) {
    sgNode *o = (sgNode *)selectionBuffer[i] ;
    
#if 0
    std::cerr << o->getName() << " < " << std::flush ;
#endif
    o->applyTransformations() ;
  }

#if 0
  std::cerr << std::endl ;
#endif

  GLdouble mvmatrix[16] ;
  glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix) ;

  glReadBuffer(GL_FRONT) ;
  //glReadBuffer(GL_BACK) ;
  glReadPixels(x,viewport[3]-y-1, 1,1, GL_DEPTH_COMPONENT,GL_FLOAT, (GLvoid *)&z) ;

  glReadBuffer(GL_BACK) ;

  GLint ret;
  /* FIXME: !! */
  if (first_time == 1)
  {
	  if (z > 0 && z <= 0.04)
	  {
		  hack = 256;
	  }
	  first_time = 0;
  }
  ret = gluUnProject(x,viewport[3]-y-1,z*hack, mvmatrix, projmatrix, viewport, ox, oy, oz) ;
#if 0
  GLint d;
  glGetIntegerv(GL_DEPTH_BITS, &d);
  std::cerr << "depth bits: " << d << std::endl ;
  std::cerr << " x=" << x << " y=" << y << " (" << viewport[3]-y-1 << ") z=" << z << std::endl ;
  std::cerr << " ox=" << *ox << " oy=" << *oy << " oz=" << *oz << " ("<< ret << ")" << std::endl ;
#endif
}

void
unproject(
	int x, int y, float z, sgViewpoint *viewpoint, GLuint *selectionBuffer,
	int selectionBufferSize, GLdouble *ox, GLdouble *oy, GLdouble *oz)
{

	GLint viewport[4] ;
	glGetIntegerv(GL_VIEWPORT, viewport) ;

	viewpoint->applyTransforms() ;

	GLdouble projmatrix[16] ;
	glGetDoublev(GL_PROJECTION_MATRIX, projmatrix) ;

	for (int i=0; i<selectionBufferSize; ++i) {
		sgNode *o = (sgNode *)selectionBuffer[i] ;
		o->applyTransformations() ;
	}

	GLdouble mvmatrix[16] ;
	glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix) ;

	glReadBuffer(GL_BACK) ;

	GLint ret;
	ret = gluUnProject(
		x,viewport[3]-y-1,z, mvmatrix, projmatrix, viewport, ox, oy, oz);
#if 0
	std::cerr << " x=" << x << " y=" << y << " (" << viewport[3]-y-1
		  << ") z=" << z << std::endl ;
	std::cerr << " ox=" << *ox << " oy=" << *oy << " oz=" << *oz
		  << " ("<< ret << ")" << std::endl ;
#endif
}

void
ASetShape(Display *dpy, Window w,
	  sgViewpoint *viewpoint, GLuint *selectionBuffer, 
	  int width, int height) {
#if HAVE_SHAPE
  Pixmap shape;
  int x,y;
  static Pixmap ip = None;
  static GC gc = None;

  if (gc == None) {
    XGCValues	xgcv;
    ip = XCreatePixmap (dpy, w, 1, 1, 1);
    gc = XCreateGC (dpy, ip, 0, &xgcv);
  }
  shape = XCreatePixmap (dpy, w, width, height, 1);
  XSetForeground (dpy, gc, 0);
  XFillRectangle (dpy, shape, gc, 0, 0, width, height);
  XSetForeground (dpy, gc, 1);

  for (y=0;y<height; y++) {
      GLfloat *segments = new GLfloat [1024] ;
      int size = viewpoint->pickXSegments(width/2, y, width, segments, 1024);
      int s;
      for(s=0; s < size; s = s+2) {
	XDrawLine(dpy, shape, gc, (int)segments[s], y, (int)segments[s+1], y);
      }
      delete [] segments;
  }
  std::cerr << "Shape" << std::endl ;

  XShapeCombineMask (dpy, w, ShapeBounding, 0, 0, shape, ShapeSet);
  XFreePixmap(dpy,shape);
#endif
}
