/*
 *
 * main/LayerManaer.H --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef LAYERMANAGER_H
#define LAYERMANAGER_H

#include <nucleo/gl/scenegraph/sgNode.H>
#include <nucleo/gl/scenegraph/sgViewpoint.H>
#include "fvwmmodule/FvwmModule.H"

#define SURFACE_SPHERE            0
#define SURFACE_TBCYLINDER        1
#define SURFACE_LRCYLINDER        2
#define SURFACE_SPHERE_CYLINDER    3
#define SURFACE_CYLINDER_SPHERE   4
#define SURFACE_LAST            SURFACE_CYLINDER_SPHERE
#define SURFACE_FIRST           SURFACE_SPHERE

#define SURFACE_DEFAULT SURFACE_SPHERE

#define SURFACE_DEFAULT_CUT 0.5

using namespace nucleo ;

//class XwncWindow;

class LayerManager : public sgNode {

 private:
	sgNode *_bg;
	sgNode *_rootWin;
	sgViewpoint *_viewpoint;

	float _depth;

	float _desktopWidth;
	float _desktopHeight;
	float _eyeDist;
	double _rootStart;
	double _bigGap;

	float _transX;
	float _transY;
	float _scaleX;
	float _scaleY;

	float _surfaceTop;
	float _surfaceLeft;
	float _surfaceBottom;
	float _surfaceRight;
	float _surfaceX;
	float _surfaceY;
	int _surface;
	bool _surfaceOn;

	bool _inExposeMode;

	void _getParameters(
		float *unmanagedDepth, float *managedDepth, int *unmanagedNum,
		int *managedNum);
	float _getDeltaDepth(
		XwncWindow *win, int *count, float unmanagedDepth,
		float managedDepth, int numUnmanaged, int numManaged);
	void _printStacking(void);
	unsigned int _guessGravity(XwncWindow *win);

 protected:
	void display(dlPolicy policy) ;
	void select(void) ;

 public:
	LayerManager(std::string name) ;
	void setBg(sgNode *bg);
	void setRootWin(sgNode *rootWin);
	XwncWindow *getRootWindow();
	void setDepth(float depth);
	float getDepth(float depth);
	void restack(sgNode *node, sgNode *nodeNext, bool add);
	XwncWindow *getUnmanagedWindow(int x, int y);
	void setDesktopProperties(
		float desktopWidth, float desktopHeight, double eyeDist,
		double near, double far, sgViewpoint *viewpoint);
	void translate_rel(float x, float y, float z);
	void getTranslation(float *x, float *y);

	// input and output are X11 geometries
	void x11projectWarp(
		int x, int y, int *px, int *py, bool warp = true);

	// translate the desktop (if scaled) so that win is visible
	void translateForWindow(XwncWindow *win);

	void scaleWindow(
		XwncWindow *win, GLfloat fx, GLfloat fy, GLfloat fz,
		unsigned int gravity = 0, bool absolute = false);
	void rotateWindow(
		XwncWindow *win, GLfloat angle, GLfloat fx, GLfloat fy,
		GLfloat fz, unsigned int gravity = 0, bool absolute = false);

	void scale(float x, float y, float z);
	void resetTransformations(void);

	void setTmpAlphaAbove(
		XwncWindow *startWindow, bool set = true, float alpha = 0.5);

	void setSurfaceParameters(
		float surfaceTop, float surfaceLeft, float _surfaceBottom,
		float _surfaceRight);
	void setSurfaceParameters(float cutX, float cutY);
	void setSurfaceParameters(float cut);
	// surface by name
	void setSurface(int surface, float cutX, float cutY);
	void setSurface(int surface, float cut);
	void setSurface(int surface);
	void setSurface(float cutX, float cutY);
	void setSurface(float cut);
	//
	void enableSurface(int toggle);
	void getSurfaceRotation(float x, float y, double *rx, double *ry);

	void exposeMode(void);
	bool inExposeMode(void);
} ;

#endif
