/*
 *
 * wncdesktop/XwncDesktop.H --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */


#ifndef _XwncDesktop_H
#define _XwncDesktop_H

#include "config.h"

#include <string>
#include <map>

#include <X11/Xlib.h>
#include <X11/Xmd.h>

#include <nucleo/gl/scenegraph/sgNode.H>
#include <nucleo/gl/window/glWindow.H>

#include "XwncWindow.H"
#include "wncSource.H"
#include "main/AUtils.H"

class AScreen;

class XwncDesktop {

private:

	std::string _style ;
	//sgNode *_parent ;

	Display *_xdisplay ;

	typedef std::map<Window, XwncWindow*> WindowMapping ;
	WindowMapping _windows ;

	wncSource *_wncServer ;
	AScreen *_ascr;
	ACursor *_cursor;

	bool _useRGBTex;

	bool _modifierPressed[256];
	
	void _addWindow(
		Window id, int x, int y, unsigned int width, unsigned int height,
		bool map, int isroot);

public:

	XwncWindow *findWindow(Window id);

	void handleConfigureWindow(
		Window id, int x, int y, int width, int width, int isroot);
	void handleCursorPosition(int x, int y);
	void handleImageFramebufferUpdate(
		XwncWindow *win, bool isRoot, Image *img, int x, int y,
		unsigned int w, unsigned int h);
	void handleWindowShape(XwncWindow *win, CARD16 *buf, int size);
	void handleUnmapWindow(Window id);
	void handleDestroyWindow(Window id);
	void handleRestackWindow(
		Window id, Window nextId, Window transientFor,
		Window unmanagedFor, unsigned long flags);

	void setSoftwareCursor(ACursor *cursor);

	void rootPointerEvent(int x, int y, unsigned long button_mask);
	void rootKeyEvent(unsigned long key, bool down_flag);
	void setModifiers(glWindow::event e, bool down_flag);
	void releaseModifiers(void);

	void getSize(unsigned int *width, unsigned int *height);

	void startRequest(void);
	void setAScreen(AScreen *ascr);

	XwncDesktop(
		std::string rfburi, std::string style, bool use_shm,
		bool use_rgb_tex);
	XwncDesktop(std::string rfburi);

	Bool HandleXCursor(
		int xhot, int yhot, int width, int height, char *buf,
		rfbXCursorColors colors);
	Bool HandleARGBCursor(
		CARD32 xhot, CARD32 yhot, CARD32 width, CARD32 height,
		XcursorPixel *buf);
} ;

#endif
