/*
 *
 * wncdesktop/XwncDesktop.cxx --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include <stdio.h>

#include <nucleo/gl/scenegraph/sgNode.H>
#include <nucleo/gl/scenegraph/sgClipper.H>
#ifndef __APPLE__
#include <nucleo/gl/window/glWindow_GLX.H>
#endif
#include "XwncDesktop.H"

#include "main/AScreen.H"
#include "wstyle/AbstractWindowRenderer.H"

#define DEBUG_LEVEL 0

//-----------------------------------------------------------------------

XwncWindow *XwncDesktop::findWindow(Window id)
{
	WindowMapping::iterator i = _windows.find(id);

	if (i !=_windows.end())
	{
		return (*i).second ;
	}

	return 0;
}

void XwncDesktop::_addWindow(
	Window id, int x, int y, unsigned int width, unsigned int height,
	bool map, int isroot)
{
	char tmp[256] ;
	sprintf(tmp,"0x%lx",id);

	XwncWindow *window = new XwncWindow(
		tmp, id, _wncServer, _ascr, x, y, width, height);

	if (!window)
	{
		std::cerr << "Unable to create XwnxWindow for " << tmp
			  << std::endl;
		return;
	}
  
	_windows[id] = window;
	if (map)
	{
		_ascr->root->addDependency(window);
		window->setMapped(true);
	}
	else
	{
		window->setMapped(false);
	}

	if (isroot)
	{
		_ascr->root->setRootWin(window);
		window->setRootWindow();
	}

	AbstractWindowRenderer *renderer = createWindowRenderer(_style, window) ;
	window->setRenderer(renderer) ;

	unsigned int win_w, win_h;
	int d_x,d_y;
	_ascr->ametista->getGeometry(&win_w, &win_h, &d_x,&d_y);
	float win_width = win_w;
	float win_height = win_h;
	float px = x;
	float py = y;
	
	window->translate_rel(
		- (win_width-width)/2 +  px, (win_height-height)/2 - py, 0);
	window->setRealPosition(x, y);
	if (isroot)
	{
#if 0
		std::cerr << "Added the root window "
			  << x << " " << y << " " << width << "x" << height
			  << " a: " << d_x << " " << d_y << " " << win_w << "x"
			  << win_h << std::endl;
#endif
	}
#if DEBUG_LEVEL > 0
	std::cerr << "Added Window: " << tmp << " at:" << px << ","
		  << py <<std::endl;
#endif
}

void XwncDesktop::handleConfigureWindow(
	Window id, int x, int y, int width, int height, int isroot)
{
	WindowMapping::iterator i = _windows.find(id);

	XwncWindow *win = findWindow(id);

	if (!win)
	{
		_addWindow(id, x, y, width, height, false, isroot);
		return;
	}
	
	float oldX, oldY;
	float oldW, oldH;
	bool changed = false;

	if (isroot)
	{
		std::cerr << "Get a config for the root window" << std::endl;
		return; // ???
	}
	win->getCurrentSize(&oldW, &oldH);
	win->getCurrentRealPosition(&oldX, &oldY);

#if 0
	std::cerr << "Get a ConfigureWindow" << std::endl;
#endif
	if ((int)oldW != width || (int)oldH != height)
	{
		win->getRenderer()->setShape(0, 0, _ascr->hasStencilBuffer());
		win->resetTexture(width, height);
		changed = true;
#if 0
		std::cerr << "New Size: " << width << " / " << oldW << " // "
			  << height << " / " << oldH << std::endl;
#endif
	}
	if (x != oldX || y != oldY)
	{

		if (win->inTextureReset())
		{
			win->textureResetSetRealPosition(x, y);
		}
		else
		{
			win->translate_rel(-oldX + x, oldY - y, 0);
			win->setRealPosition(x, y);
		}
		changed = true;
#if 0
		std::cerr << "New Positon: " << x << " / " << oldX
			  << " // " << y << " / " << oldY << std::endl;
#endif
	}

	if (_ascr->waitingForAConfigureWindow())
	{
		_ascr->draw(false);
		_ascr->resetSelection();
	}
	else if (changed && !_ascr->inInteractiveMove() &&
		 !_ascr->inInteractiveResize())
	{
		_ascr->scheduleResetSelection();
		//_ascr->draw(false);
		//_ascr->resetSelection();
	}
}

void
XwncDesktop::handleCursorPosition(int x, int y) {
  int px,py;
  _ascr->root->x11projectWarp(x, y, &px, &py);
#if 0
  std::cerr << "XwncDesktop::handleCursorPosition: " << x << "," << y << " --> " << px << "," << py << std::endl;
#endif
  _ascr->ametista->warpCursor(px, py) ;
  if (!_ascr->inInteractiveMove() && !_ascr->inInteractiveResize()) {
    _ascr->resetSelection(px, py);
  } else {
    // FIXME (moving/resizing with the keyboard arrows)
  }

}

void XwncDesktop::handleImageFramebufferUpdate(
	XwncWindow *win, bool isRoot, Image *img, int x, int y, unsigned int w,
	unsigned int h)
{

	if (!win)
	{
		std::cerr << "handleImageFramebufferUpdate for a window we do "
			  << "not know" << std::endl ;
		return;
	}

	float wH, wW;
	win->getCurrentSize(&wW, &wH);
	if (x+w > (unsigned int)wW || y+h > (unsigned int)wH)
	{
		return;
	}
	win->updateTexture(img, x, y, w, h);
}

void XwncDesktop::handleWindowShape(XwncWindow *win, CARD16 *buf, int nrecs)
{
	if (!win)
	{
		std::cerr << "handleWindowShape for a window we do "
			  << "not know" << std::endl ;
		if (buf)
		{
			free(buf);
		}
		return;
	}

	win->getRenderer()->setShape(buf, nrecs, _ascr->hasStencilBuffer());

	if (_ascr->hasStencilBuffer())
	{
		return;
	}

	if (_useRGBTex)
	{
		// FIXME
		return;
	}
	// FIXME

}

void XwncDesktop::handleUnmapWindow(Window id)
{
	WindowMapping::iterator i = _windows.find(id);

	if (i == _windows.end())
	{
		std::cerr << "Trying to unmap a window we don't know ("
			  << std::hex << id << std::dec << ")" << std::endl ;
		return;
	}

	XwncWindow *win = (*i).second ;

	if (!win->isMapped())
	{
		return;
	}

	win->setMapped(false);

	_ascr->root->removeDependency(win);
	win->unmapCorrection();
	_ascr->draw(false);
	_ascr->resetSelection();
	// save memory. we must get a full frame buffer update at remap
	win->deleteTexture();
}

void XwncDesktop::handleDestroyWindow(Window id)
{
	WindowMapping::iterator i = _windows.find(id) ;
	if (i==_windows.end())
	{
		std::cerr << "Trying to remove a window we don't know ("
			  << std::hex << id << std::dec << ")" << std::endl ;
		return ;
	}

	XwncWindow *win = (*i).second ;

	_windows.erase(id) ;
	if (win->isMapped())
	{
		_ascr->root->removeDependency(win) ;
		_ascr->draw(false);
		_ascr->resetSelection();
	}

#if DEBUG_LEVEL > 0
	std::cerr << std::hex << id << std::dec << " removed from the desktop"
		  << std::endl ;
#endif

	delete win;
}

void XwncDesktop::handleRestackWindow(
	Window id, Window nextId, Window transientFor, 
	Window unmanagedFor, unsigned long flags)
{
	WindowMapping::iterator i = _windows.find(id);
	WindowMapping::iterator j = _windows.find(nextId);

	if (i == _windows.end())
	{
		std::cerr << "Trying to restack a window we don't know ("
			  << std::hex << id << std::dec << ")" << std::endl ;
		return;
	}

	XwncWindow *win = (*i).second ;

	XwncWindow *nextWin = 0;
	if (j != _windows.end())
	{
		nextWin = (*j).second ;
	}


	if ((flags & rfbWindowFlagsOverrideRedirect) ||
	    (flags & rfbWindowFlagsUnmanaged))
	{
		win->setUnmanaged();
	}

	if (flags & rfbWindowFlagsEwmhDesktop)
	{
		win->setEwmhDesktop();
	}
	if (flags & rfbWindowFlagsNetChecking)
	{
		win->hide();
	}
	if (flags & rfbWindowFlagsTransient)
	{
		win->setTransientFor(transientFor);
#if 0
		std::cerr << "Transient For " << std::hex <<  transientFor
			  << std::dec << std::endl;
#endif
	}
	if (unmanagedFor != None)
	{
		win->setUnmanagedFor(unmanagedFor);
#if 0
		std::cerr << "Unmanaged For " << std::hex <<  unmanagedFor
			  << std::dec << std::endl;
#endif
	}
	if (flags & rfbWindowFlagsToolTips)
	{
#if 0
		std::cerr << "ToolTips " << std::endl;
#endif
		win->setToolTips();
	}


	if (flags & rfbWindowFlagsInputOnly)
	{
		std::cerr << "Input Only window " << std::hex << id << std::dec
			  << std::endl;
	}

	if (win->isMapped())
	{
		_ascr->root->restack(win, nextWin, false);
		//_ascr->scheduleResetSelection(); <- more is needed!
		_ascr->draw(false);
		_ascr->resetSelection();
	}
	else
	{
		win->mapCorrection();
		_ascr->root->restack(win, nextWin, true);
		if (!win->isUnmanaged() && win->neverMapped())
			_ascr->root->translateForWindow(win);
		win->setMapped(true);
		_ascr->draw(false);
		_ascr->resetSelection();
	}
}
// ----------------------------------------------------------------------------
void 
XwncDesktop::setSoftwareCursor(ACursor *cursor)
{
	_cursor = cursor;
}

// ----------------------------------------------------------------------------

void 
XwncDesktop::rootPointerEvent(int x, int y, unsigned long button_mask) {
  _wncServer->pointerEvent(DefaultRootWindow(_xdisplay),
					  x, y,
					  button_mask);
}

void 
XwncDesktop::rootKeyEvent(unsigned long key, bool down_flag) {
	_wncServer->keyEvent(key, down_flag);
}

void 
XwncDesktop::setModifiers(glWindow::event e, bool down_flag) {
  if (IsModifierKey(e.keysym)) {
    KeyCode kc = XKeysymToKeycode(_xdisplay, e.keysym);
    if ((int)kc < 256) _modifierPressed[(int)kc] = down_flag;
  }
}

void 
XwncDesktop::releaseModifiers(void) {
	int i;
	KeySym ks;

	for (i = 0; i < 256; i++)
	{
		if (_modifierPressed[i])
		{
			ks = XKeycodeToKeysym(_xdisplay, i, 0);
			_wncServer->keyEvent(ks, false);
			_modifierPressed[i] = false;
		}
	}
}

void XwncDesktop::getSize(unsigned int *width, unsigned int *height)
{
	_wncServer->getSize(width, height);
}

// -------------------------------------------------------------------------
void XwncDesktop::startRequest(void)
{
	_wncServer->startRequest();
}

void XwncDesktop::setAScreen(AScreen *ascr)
{
	_ascr = ascr;
}

XwncDesktop::XwncDesktop(
	std::string rfburi, std::string style, bool use_shm, bool use_rgb_tex)
{
	_style = style ;
	_xdisplay = 0 ;
	_useRGBTex = use_rgb_tex;
	_cursor = 0;

	URI uri(rfburi) ;

	std::string vncdisplay =
		uri.getHost() + std::string(":") + uri.getPort() ;
	_xdisplay = XOpenDisplay(vncdisplay.c_str());

	if (!_xdisplay)
	{
		std::cerr << "[XwncDesktop] Can't open VNC display"
			  << std::endl ;
		exit(1);
	}

	_wncServer = new wncSource(this, use_rgb_tex, uri, use_shm);
	if (!_wncServer)
	{
		std::cerr << "[XwncDesktop] Can't create wncSource"
			  << std::endl ;
		exit(1);
	}

	bool ret = _wncServer->start();

	if (!ret)
	{
		std::cerr << "[XwncDesktop] Can't start wncSource" << std::endl;
		exit(1);
	}

	for (int i = 0; i < 256; i++)
	{
		_modifierPressed[i] = false;
	}
}

XwncDesktop::XwncDesktop(std::string rfburi)
{
	XwncDesktop(rfburi, "foldable", false, false);
}

// --------------------------------------------------------------------------
// react ...

// --------------------------------------------------------------------------

/* Copied from Xvnc/lib/font/util/utilbitmap.c */
static unsigned char _reverse_byte[0x100] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};

Bool
XwncDesktop::HandleXCursor(
	int xhot, int yhot, int width, int height, char *buf,
	rfbXCursorColors colors)
{
	size_t bytesPerRow, bytesData;

	bytesPerRow = (width + 7) / 8;
	bytesData = bytesPerRow * height;

	for (unsigned int i = 0; i < bytesData * 2; i++)
	{
		buf[i] = (char)_reverse_byte[(int)buf[i] & 0xFF];
	}

	if (_cursor)
	{
		_cursor->handleXCursor(
			xhot, yhot, width, height, buf, colors);
		return true;
	}
#ifdef __APPLE__
	// _cursor != 0
#else
	XColor bg, fg;
	Drawable dr;
	Display *dpy;
	unsigned int wret = 0, hret = 0;
	Pixmap source, mask;
	Cursor cursor;
	static Bool prevXCursorSet = False;
	static Cursor prevXCursor;
	glWindow_GLX *castedWin;

	castedWin = dynamic_cast<glWindow_GLX *>(_ascr->ametista);
	if (castedWin == NULL)
	{
		std::cerr <<"XwncDesktop::HandleXCursor GLX cast fail! ("
			  << __FILE__ << ", " << __LINE__ << ")" << std::endl;
	}
	dpy = (Display*)castedWin->getDisplay();
	dr = DefaultRootWindow(dpy);

	if (width * height)
	{
		XQueryBestCursor(dpy, dr, width, height, &wret, &hret);
	}

	if (wret < (unsigned int)width || hret < (unsigned int)height)
	{
		// Switch to software cursor! (FIXME)
		std::cerr << "Metisse switch to software cursor!" << std::endl;
		if (_cursor == 0)
		{
			_ascr->showCursor();
		}
		_cursor->handleXCursor(
			xhot, yhot, width, height, buf, colors);
		return true;
	}

	if (buf == NULL || width * height == 0)
	{
		/* Free resources */
		if (buf != NULL)
		{
			free(buf);
		}
		width = height = 1;
		buf = (char *)malloc(2);
		buf[0] = buf[1] = 0;
		bytesPerRow = 1;
		bytesData = 1;
	}
	if (xhot >= width)
	{
		xhot = width - 1;
	}
	else if (xhot < 0)
	{
		xhot = 0;
	}
	if (yhot > height)
	{
		yhot = height - 1;
	}
	else if (yhot < 0)
	{
		yhot = 0;
	}

	bg.red   = (unsigned short)colors.backRed   << 8 | colors.backRed;
	bg.green = (unsigned short)colors.backGreen << 8 | colors.backGreen;
	bg.blue  = (unsigned short)colors.backBlue  << 8 | colors.backBlue;
	fg.red   = (unsigned short)colors.foreRed   << 8 | colors.foreRed;
	fg.green = (unsigned short)colors.foreGreen << 8 | colors.foreGreen;
	fg.blue  = (unsigned short)colors.foreBlue  << 8 | colors.foreBlue;

	source = XCreateBitmapFromData(dpy, dr, buf, width, height);
	mask = XCreateBitmapFromData(
		dpy, dr, &buf[bytesData], width, height);
	cursor = XCreatePixmapCursor(
		dpy, source, mask, &fg, &bg, xhot, yhot);
	XFreePixmap(dpy, source);
	XFreePixmap(dpy, mask);
	free(buf);

	if (castedWin->getWindowID())
	{
		XDefineCursor(dpy, castedWin->getWindowID(), cursor);
	}
	if (prevXCursorSet)
	{
		XFreeCursor(dpy, prevXCursor);
		prevXCursorSet = False;
	}
	prevXCursor = cursor;
	prevXCursorSet = True;

	return True;
#endif
}

Bool
XwncDesktop::HandleARGBCursor(
	CARD32 xhot, CARD32 yhot, CARD32 width, CARD32 height, XcursorPixel *buf)
{

	if (_cursor)
	{
		_cursor->handleARGBCursor(xhot, yhot, width, height, buf);
		return true;
	}
#ifdef __APPLE__
	// _cursor != 0
#else
#ifdef HAVE_XCURSOR
	static Bool prevXCursorSet = False;
	static Cursor prevXCursor;
	Cursor cursor;
	glWindow_GLX *castedWin;
	Display *dpy;
	XcursorImage *image;

	castedWin = dynamic_cast<glWindow_GLX *>(_ascr->ametista);
	if (castedWin == NULL)
	{
		std::cerr <<"XwncDesktop::HandleXCursor GLX cast fail! ("
			  << __FILE__ << ", " << __LINE__ << ")" << std::endl;
	}
	dpy = (Display*)castedWin->getDisplay();

	image = XcursorImageCreate (width, height);

	image->xhot = xhot;	/* hot spot x (must be inside image) */
	image->yhot = yhot;	/* hot spot y (must be inside image) */
	image->pixels = buf;	/* pointer to pixels */

	cursor = XcursorImageLoadCursor(dpy, image);

	if (castedWin->getWindowID())
	{
		XDefineCursor(dpy, castedWin->getWindowID(), cursor);
	}
	if (prevXCursorSet)
	{
		XFreeCursor(dpy, prevXCursor);
		prevXCursorSet = False;
	}
	prevXCursor = cursor;
	prevXCursorSet = True;

	if (buf)
	{
		free(buf);
		image->pixels = NULL;
	}

	if (image)
	{
		XcursorImageDestroy(image); 
	}

	return True;
#endif
#endif
}
