/*
 *
 * wncdesktop/XwncWindow.H --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */


#ifndef _XwncWindow_H
#define _XwncWindow_H


#include <string>

#include <nucleo/gl/scenegraph/sgNode.H>

#include "texture/glTiledTexturedImage.H"
#include "wstyle/AbstractWindowRenderer.H"
#include "wncSource.H"

class AScreen;

class XwncWindow : public sgNode {

 private:

	std::string _title ;
	wncSource *_wncServer ;
	AScreen *_ascr;

	bool _resetTexture;
	float _rwidth;
	float _rheight;
	float _trx;
	float _try;
	float _tx;
	float _ty;
	float _rrx;
	float _rry;

	glTiledTexturedImage *_texture;

	int _texCoordsSize ;
	float *_texCoords ;

	// X11 geometry
	float _real_x;
	float _real_y;
	float _width;
	float _height;

	// additional translation vs the X11 translation
	float _interTransX;
	float _interTransY;

	unsigned long _frameID;

	bool _mapped;
	bool _neverMapped;
	bool _changed;

	float _red, _green, _blue, _alpha ;
	float _alpha_saved;

	AbstractWindowRenderer* _renderer;
	
	bool _unmanaged;
	bool _isRootWindow;
	bool _isEwmhDesktop;
	bool _isToolTips;
	unsigned long _transientFor;
  	unsigned long _unmanagedFor;
	bool _hide;

	float _layerCorrection;
	float _desktopWidth;
	float _desktopHeight;
	float _eyeDist;

	float _rootTransY;
	float _rootTransX;

	double _surfaceRotationY;
	double _surfaceRotationX;
	bool _surfaceRotationAllowed;
	void _surfaceCorrection(float tx, float ty);
	void _resetSurfaceCorrection(void);
	void _translate_rel2(float x, float y, float z);

	float _absoluteScaleX;
	float _absoluteScaleY;
	float _absoluteScaleZ;

	float _absoluteRotateAngle;
	float _absoluteRotateX;
	float _absoluteRotateY;
	float _absoluteRotateZ;

 public:

	XwncWindow(
		std::string title, unsigned long id, wncSource *wncServer,
		AScreen *ascr, int x, int y, unsigned int width,
		unsigned int height);
	~XwncWindow();

	static float rootTransY;
	static float rootTransX;

	std::string getTitle(void);
	
	void setRenderer(AbstractWindowRenderer *renderer);
	AbstractWindowRenderer *getRenderer(void);
	
	void setSize(float width, float height);
	void getCurrentSize(float *width, float *height);
	void getSize(int *width, int *height);
	void getSize(float *width, float *height);
	void getSize(double *width, double *height);
	void setRealPosition(float x, float y);
	void getCurrentRealPosition(float *x, float *y);
	void getRealPosition(float *x, float *y);
	void getRealPosition(int *x, int *y);
	void setMapped(bool mapped);
	bool isMapped(void);
	bool neverMapped(void);
	void setUnmanaged(bool u);
	void setUnmanaged(void);
	bool isUnmanaged(void);
	void setRootWindow(void);
	bool isRootWindow(void);
	void setEwmhDesktop(void);
	bool isEwmhDesktop(void);
	void setToolTips(void);
	bool isToolTips(void);
	void setTransientFor(unsigned long w);
  	void setUnmanagedFor(unsigned long w);
	unsigned long  getTransientFor(void);
  	unsigned long  getUnmanagedFor(void);
	void hide(void);

	unsigned long getFrameID(void);

	void setBgcolor(float red, float green, float blue, float alpha);
	void setAlpha(float alpha);
	void setTmpAlpha(float alpha);
	void resetAlpha(void);
	void getBgcolor(float *red, float *green, float *blue, float *alpha);
	float getAlpha(void);
	bool doPseudoSelection(void);

	void resetTexture(void);
	bool inTextureReset(void);
	void textureResetSetRealPosition(float x, float y);
	void resetTexture(float width, float height);
	void deleteTexture(void);

	void updateTexture(
		Image *img, int x, int y, unsigned int w, unsigned int h);

	void surfaceRotation(bool cancel);
	void layerCorrection(
		float desktopWidth, float desktopHeight, float eyeDist,
		float depth);
	void restoreTranslation(void);
	void mapCorrection(void);
	void unmapCorrection(void);
	void internalTranslation(float x, float y, float z);
	void translate_rel(float x, float y, float z);
	void absoluteScale(float x, float y, float z);
	void absoluteRotate(GLfloat angle, GLfloat x, GLfloat y, GLfloat z);

	void getTextureMapping(
		glTiledTexturedImage **texture, int *texCoordsSize,
		float **texCoords);

	void setTextureCoordinates(
		float left, float top, float right, float bottom);

	void keyEvent(unsigned long key, bool down_flag);
	void pointerEvent(int x, int y, unsigned char button_mask);

};

#endif
