/*
 *
 * wncdesktop/XwncWindow.cxx --
 *
 * Copyright (C) Nicolas Roussel
 * Copyright (C) Olivier Chapuis
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include <nucleo/image/sink/ImageSink.H>
#include <nucleo/gl/glUtils.H>

#include "main/AScreen.H"

#include "wstyle/AbstractWindowRenderer.H"

#include "wncSource.H"
#include "XwncDesktop.H"
#include "XwncWindow.H"

float XwncWindow::rootTransY = 0;
float XwncWindow::rootTransX = 0;

XwncWindow::XwncWindow(
	std::string title, unsigned long id, wncSource *wncServer,
	AScreen *ascr, int x, int y, unsigned int width, unsigned int height)
{
  
	_title = title;
	_wncServer = wncServer;
	_ascr = ascr;
	_real_x = _rrx = x;
	_real_y = _rry = y;
	x = y = 0;
	_interTransX = _interTransY = 0;

	_width = _rwidth = width;
	_height = _rheight = height;
	_frameID = id;

	_texture = (glTiledTexturedImage*)0;

	_resetTexture = false;
	_trx = 0;
	_try = 0;

	_mapped = false;
	_neverMapped = true;
	_changed = true;

	_red = _green = _blue = 1.0;
	_alpha = _alpha_saved = 1.0;
 
	_texCoordsSize = 0 ;
	_texCoords = (float*)0 ;
	_renderer = (AbstractWindowRenderer*)0 ;

	_unmanaged = false;
	_isRootWindow = false;
	_isEwmhDesktop = false;
	_isToolTips = false;
	_transientFor = 0;
	_unmanagedFor = 0;
	_hide = 0;

	_layerCorrection = 0;
	_desktopWidth = 0;
	_desktopHeight = 0;
	_eyeDist = 0;

	_rootTransX = 0;
	_rootTransY = 0;

	_surfaceRotationY = 0;
	_surfaceRotationX = 0;
	_surfaceRotationAllowed = true;

	_absoluteScaleX = 1;
	_absoluteScaleY = 1;
	_absoluteScaleZ = 1;

	_absoluteRotateAngle = 0;
	_absoluteRotateX = 0;
	_absoluteRotateY = 0;
	_absoluteRotateZ = 0;
}

XwncWindow::~XwncWindow()
{
	if (_texCoords)
	{
		delete [] _texCoords;
	}
	if (_texture)
	{
		delete _texture;
	}
	if (_renderer)
	{
		delete _renderer;
	}
	
}

// --------------------------------------------------------------

std::string
XwncWindow::getTitle(void)
{
	return _title ;
}

void
XwncWindow::setRenderer(AbstractWindowRenderer *renderer)
{
	if (_renderer)
	{
		removeDependency(_renderer) ;
		delete _renderer ;
	}
	_renderer = renderer ;
	addDependency(renderer) ;
	_changed = true ;
}

AbstractWindowRenderer *XwncWindow::getRenderer(void)
{
	return _renderer;
}

// --------------------------------------------------------------

void XwncWindow::setSize(float width, float height)
{
	_width = _rwidth = width ;
	_height = _rheight = height ;
	_changed = true;
}

void XwncWindow::getCurrentSize(float *width, float *height)
{
	*width = (float)_rwidth ;
	*height = (float)_rheight ;
}

void XwncWindow::getSize(float *width, float *height)
{
	*width = (float)_width ;
	*height = (float)_height ;
}

void XwncWindow::getSize(double *width, double *height)
{
	*width = (float)_width ;
	*height = (float)_height ;
}

void XwncWindow::getSize(int *width, int *height)
{
	*width = (int)_width ;
	*height = (int)_height ;
}

void XwncWindow::setRealPosition(float x, float y)
{
	_real_x = _rrx = x;
	_real_y = _rry = y;
}

void XwncWindow::getCurrentRealPosition(float *x, float *y)
{
	*x = _rrx;
	*y = _rry;
}

void XwncWindow::getRealPosition(float *x, float *y)
{
	*x = _real_x;
	*y = _real_y;
}

void XwncWindow::getRealPosition(int *x, int *y)
{
	*x = (int)_real_x;
	*y = (int)_real_y;
}
void XwncWindow::setMapped(bool mapped)
{
	_mapped = mapped;
	if (_mapped)
		_neverMapped = false;
}


bool XwncWindow::isMapped(void)
{
	return _mapped;
}

bool XwncWindow::neverMapped(void)
{
	return _neverMapped;
}
void XwncWindow::setUnmanaged(bool u)
{
	_unmanaged = u;
}

void XwncWindow::setUnmanaged(void)
{
	setUnmanaged(true);
}

bool XwncWindow::isUnmanaged(void)
{
	return _unmanaged;
}

void XwncWindow::setRootWindow(void)
{
	_isRootWindow = true;
}

bool XwncWindow::isRootWindow(void)
{
	return _isRootWindow;
}

void XwncWindow::setEwmhDesktop(void)
{
	_isEwmhDesktop = true;
}

bool XwncWindow::isEwmhDesktop(void)
{
	return _isEwmhDesktop;
}

void XwncWindow::setToolTips(void)
{
	_alpha = 0.8; // FIXME: should be configurable
	_isToolTips = true;
}

bool XwncWindow::isToolTips(void)
{
	return _isToolTips;
}

void XwncWindow::setTransientFor(unsigned long w)
{
	_transientFor = w;
}

unsigned long  XwncWindow::getTransientFor(void)
{
	return _transientFor;
}

void XwncWindow::setUnmanagedFor(unsigned long w)
{
	_unmanagedFor = w;
}

unsigned long  XwncWindow::getUnmanagedFor(void)
{
	return _unmanagedFor;
}

void XwncWindow::hide(void)
{
	_hide = true;
}

unsigned long XwncWindow::getFrameID(void)
{
	return _frameID;
}

// --------------------------------------------------------------

void XwncWindow::setBgcolor(float red, float green, float blue, float alpha)
{
	_red = red ;
	_green = green ;
	_blue = blue ;
	_alpha = alpha ;
	_changed = true ;
}

void XwncWindow::setAlpha(float alpha)
{
	_alpha = alpha ;
	_changed = true ;
}

void XwncWindow::setTmpAlpha(float alpha)
{
	_alpha_saved = _alpha;
	_alpha = alpha ;
	_changed = true ;
}

void XwncWindow::resetAlpha(void)
{
	_alpha = _alpha_saved;
	_changed = true ;
}

void XwncWindow::getBgcolor(float *red, float *green, float *blue, float *alpha)
{
	*red = _red ;
	*green = _green ;
	*blue = _blue ;
	*alpha = _alpha ;
}

float XwncWindow::getAlpha(void)
{
	return _alpha ;
}

bool XwncWindow::doPseudoSelection(void)
{
	return _ascr->doPseudoSelection();
}

// --------------------------------------------------------------
// Drawing stuff

bool XwncWindow::inTextureReset(void)
{
	return _resetTexture;
}

void XwncWindow::textureResetSetRealPosition(float x, float y)
{
	_trx = _trx - _rrx + x;
	_try = _try + _rry - y;
	_rrx = x;
	_rry = y;
}

void XwncWindow::resetTexture(float width, float height)
{
	_resetTexture = true;

	_trx = _trx + (-_rwidth + width)/2;
	_try = _try + (_rheight - height)/2;

	_rwidth = width;
	_rheight = height;
}

void XwncWindow::deleteTexture(void)
{
	delete _texture;
	_texture = 0;
}

void XwncWindow::updateTexture(
	Image *img, int x, int y, unsigned int w, unsigned int h)
{

	if (_texture == 0 || _resetTexture)
	{
		Image tmpImg;
		bool badSize = false;

		if (x == 0 && y == 0 && w == _rwidth && h == _rheight)
		{
			// ok
		}
		else
		{
#if 0
		  std::cerr << "BS: " << w << " " << h << ", "  << (int)_rwidth
			    << " " << (int)_rheight << std::endl;
#endif
			tmpImg.prepareFor(
				(int)_rwidth, (int)_rheight, img->getEncoding());
			badSize = true;
		}

		if (w == 0 || h == 0)
		{
			return;
		}

		if (_resetTexture)
		{
			translate_rel(_trx, _try, 0);
		}
		_width = _rwidth;
		_height = _rheight;
		_real_x = _rrx;
		_real_y = _rry;
		_trx = 0;
		_try = 0;
		_resetTexture = false;
		
		if (_renderer)
		{
			_renderer->simpleReset();
		}

		if (_texture == 0)
		{
			_texture = new glTiledTexturedImage(512);
		}

		if (badSize)
		{
			_texture->update(&tmpImg);
			_texture->subUpdate(img, x, y, w, h);
		}
		else
		{
			_texture->update(img);
		}

		float left = 0 ;
		float top = 0 ;
		float right = _width /(float)_texture->getWidth();
		float bottom = _height /(float)_texture->getHeight();
		setTextureCoordinates(left, top, right, bottom);

		postRedisplay();
		return;
	}

	_texture->subUpdate(img, x, y, w, h);
	postRedisplay() ;

	return;
}

void XwncWindow::setTextureCoordinates(
	float left, float top, float right, float bottom)
{
	_texCoordsSize = 8;
	if (_texCoords == 0)
	{
		_texCoords = new float [_texCoordsSize] ;
	}
	_texCoords[0] = left ; _texCoords[1] = bottom ;  // bottom left
	_texCoords[2] = right ; _texCoords[3] = bottom ; // bottom right
	_texCoords[4] = right ; _texCoords[5] = top ;    // top right
	_texCoords[6] = left ; _texCoords[7] = top ;     // top left
	_changed = true ;
}

void XwncWindow::getTextureMapping(
	glTiledTexturedImage **texture, int *texCoordsSize, float **texCoords)
{
	*texture = _texture ;
	*texCoordsSize = _texCoordsSize ;
	*texCoords = new float [_texCoordsSize] ;
	memcpy(*texCoords, _texCoords, _texCoordsSize*sizeof(float)) ;
}

// --------------------------------------------------------------
#define ROTATE_REL(a,x,y,z) getRenderer()->rotate_rel(a, x, y, z)
//#define ROTATE_REL(a,x,y,z) _ascr->root->rotateWindow(this, a, x, y, z)

void XwncWindow::surfaceRotation(bool cancel)
{
	_surfaceRotationAllowed = !cancel;
	if (cancel)
	{
		ROTATE_REL(-_surfaceRotationX, 0, -1, 0);
		ROTATE_REL(-_surfaceRotationY, 1, 0, 0);
		_surfaceRotationY = 0;
		_surfaceRotationX = 0;
	}
	else
	{
		float tx = - (_desktopWidth - _width)/2 + _real_x + rootTransX
			+ _interTransX;
		float ty = (_desktopHeight - _height)/2 - _real_y + rootTransY
			+ _interTransY;
		_surfaceCorrection(tx, ty);
	}
}

void XwncWindow::_resetSurfaceCorrection(void)
{
	if (_surfaceRotationX)
		ROTATE_REL(-_surfaceRotationX, 0, -1, 0);
	if (_surfaceRotationY)
		ROTATE_REL(-_surfaceRotationY, 1, 0, 0);
	_surfaceRotationY = 0;
	_surfaceRotationX = 0;
}

void XwncWindow::_surfaceCorrection(float tx, float ty)
{
	if (isUnmanaged() || isRootWindow() || !_surfaceRotationAllowed)
	{
		return;
	}

	double rx,ry;
	_ascr->root->getSurfaceRotation(tx, ty, &rx, &ry);

	if (_surfaceRotationX)
		ROTATE_REL(_surfaceRotationX, 0, 1, 0);
	if (_surfaceRotationY)
		ROTATE_REL(_surfaceRotationY, -1, 0, 0);
	if (ry)
		ROTATE_REL(ry, 1, 0, 0);
	if (rx)
		ROTATE_REL(rx, 0, -1, 0);
	_surfaceRotationY = ry;
	_surfaceRotationX = rx;
}

// --------------------------------------------------------------

void XwncWindow::layerCorrection(
	float desktopWidth, float desktopHeight, float eyeDist, float depth)
{
	_desktopWidth = desktopWidth;
	_desktopHeight = desktopHeight;
	_eyeDist = eyeDist;
	if (_eyeDist == 0)
	{
		return;
	}
	/* current translation */
	float tx = - (_desktopWidth - _width)/2 + _real_x + rootTransX
		+ _interTransX;
	float ty = (_desktopHeight - _height)/2 - _real_y + rootTransY
		+ _interTransY;
	if (_layerCorrection != depth)
	{
		float s =  (_eyeDist - _layerCorrection) / _eyeDist;
		float u = 1/s;

		_resetSurfaceCorrection();
		getRenderer()->translate_rel(-tx*s, -ty*s, -_layerCorrection);
		getRenderer()->scale(u, u, 1);

		_layerCorrection = depth;
		s = (_eyeDist - _layerCorrection) / _eyeDist;
		getRenderer()->translate_rel(0, 0, _layerCorrection);  
		getRenderer()->scale(s, s, 1);
		getRenderer()->translate_rel(tx*s, ty*s, 0);
		_surfaceCorrection(tx, ty);
	}
	else
	{
		_surfaceCorrection(tx, ty);
	}
}

void XwncWindow::restoreTranslation(void)
{
	float s = 1;
	_interTransX = 0;
	_interTransY = 0;
	_surfaceRotationX = 0;
	_surfaceRotationY = 0;
	_absoluteScaleX = 1;
	_absoluteScaleY = 1;
	_absoluteScaleZ = 1;

	float tx = - (_desktopWidth - (float)_width)/2 +  _real_x + rootTransX;
	float ty = (_desktopHeight - (float)_height)/2 - _real_y + rootTransY;
	if (_eyeDist == 0)
	{
		getRenderer()->translate_rel(tx, ty, 0);
		return;
	}
	else
	{
		s = (_eyeDist - _layerCorrection) / _eyeDist;
	}
	getRenderer()->translate_rel(0, 0, _layerCorrection);
	getRenderer()->scale(s, s, 1);
	getRenderer()->translate_rel(tx*s, ty*s, 0);
	_surfaceCorrection(tx, ty);
}

void XwncWindow::translate_rel(float x, float y, float z)
{
	if (_layerCorrection == 0 || _eyeDist == 0)
	{
		getRenderer()->translate_rel(x, y, z);
		return;
	}
	float tx = - (_desktopWidth - _width)/2 +  _real_x + _interTransX;
	float ty = (_desktopHeight - _height)/2 - _real_y + _interTransY;
	float s = (_eyeDist - _layerCorrection) / _eyeDist;
	float u = 1/s;

	_resetSurfaceCorrection();
	getRenderer()->translate_rel(-tx, -ty, -_layerCorrection);
	getRenderer()->scale(u, u, 1);

	getRenderer()->translate_rel(s*x, s*y, 0);

	getRenderer()->translate_rel(0, 0, _layerCorrection);
	getRenderer()->scale(s, s, 1);
	getRenderer()->translate_rel(tx, ty, 0);
	float ntx = - (_desktopWidth - _width)/2 +  _real_x + x + _interTransX;
	float nty = (_desktopHeight - _height)/2 - (_real_y + y) + _interTransY;
	_surfaceCorrection(ntx, nty);
}

void XwncWindow::_translate_rel2(float x, float y, float z)
{
	if (_layerCorrection == 0 || _eyeDist == 0)
	{
		getRenderer()->translate_rel(x, y, z);
		return;
	}
	float tx = - (_desktopWidth - _width)/2 +  _real_x + _interTransX;
	float ty = (_desktopHeight - _height)/2 - _real_y + _interTransY;
	float s = (_eyeDist - _layerCorrection) / _eyeDist;
	float u = 1/s;

	getRenderer()->translate_rel(-tx, -ty, -_layerCorrection);
	getRenderer()->scale(u, u, 1);

	getRenderer()->translate_rel(s*x, s*y, 0);

	getRenderer()->translate_rel(0, 0, _layerCorrection);
	getRenderer()->scale(s, s, 1);
	getRenderer()->translate_rel(tx, ty, 0);
	/*
	float ntx = - (_desktopWidth - _width)/2 +  _real_x + x + _interTransX;
	float nty = (_desktopHeight - _height)/2 - (_real_y + y) + _interTransY;
	*/
}

void XwncWindow::internalTranslation(float x, float y, float z)
{
	//return;
	_resetSurfaceCorrection();
	_translate_rel2(-_interTransX, -_interTransY, 0);
	x = _interTransX + x;
	y = _interTransY + y;
	_interTransX = _interTransY = 0;
	_translate_rel2(x, y, z);
	_interTransX = x;
	_interTransY = y;
	float ntx = - (_desktopWidth - _width)/2 +  _real_x + _interTransX;
	float nty = (_desktopHeight - _height)/2 - _real_y + _interTransY;
	_surfaceCorrection(ntx, nty);
}

void XwncWindow::mapCorrection(void)
{
	translate_rel(
		rootTransX - _rootTransX, rootTransY - _rootTransY, 0);
}

void XwncWindow::unmapCorrection(void)
{
	_rootTransX = rootTransX;
	_rootTransY = rootTransY;
}


void XwncWindow::absoluteScale(float x, float y, float z)
{
	_resetSurfaceCorrection();
	getRenderer()->scale(
		1/_absoluteScaleX,1/_absoluteScaleY,1/_absoluteScaleZ);
	if (_absoluteScaleX == 0 || _absoluteScaleY == 0 || _absoluteScaleZ == 0)
	{
		_absoluteScaleX = 1;
		_absoluteScaleY = 1;
		_absoluteScaleZ = 1;
	}
	else
	{
		_absoluteScaleX = x;
		_absoluteScaleY = y;
		_absoluteScaleZ = 1;
	}
	getRenderer()->scale(_absoluteScaleX,_absoluteScaleY,_absoluteScaleZ);
	float ntx = - (_desktopWidth - _width)/2 +  _real_x + _interTransX;
	float nty = (_desktopHeight - _height)/2 - _real_y + _interTransY;
	_surfaceCorrection(ntx, nty);
}

void XwncWindow::absoluteRotate(GLfloat angle, GLfloat x, GLfloat y, GLfloat z)
{
	getRenderer()->rotate_rel(
		-_absoluteRotateAngle, _absoluteRotateX, _absoluteRotateY,
		_absoluteRotateZ);
	_absoluteRotateAngle = angle;
	_absoluteRotateX = x;
	_absoluteRotateY = y;
	_absoluteRotateZ = z;
	getRenderer()->rotate_rel(
		_absoluteRotateAngle, _absoluteRotateX, _absoluteRotateY,
		_absoluteRotateZ);
}

// --------------------------------------------------------------

void XwncWindow::keyEvent(unsigned long key, bool down_flag)
{
	//std::cerr << "key: " << key << " flag: " << down_flag << std::endl ;
	_wncServer->keyEvent(key, down_flag) ;
}

void XwncWindow::pointerEvent(int x, int y, unsigned char button_mask)
{
 	int realx = (int)(x + _width/2.0 + _real_x);
	int realy = (int)(-y + _height/2.0 + _real_y) - 1;

     
#if 0
	if (realx < _real_x || realx > _real_x + _width ||
	     realy < _real_y || realy > _real_y + _height)
	  {
		  // OUTCH
		  std::cerr << (int)button_mask << " : " << x << "," << y <<
			  " - " << realx << "," << realy << std::endl ;
	  }
#endif

	_wncServer->pointerEvent(
		getFrameID(), realx, realy, button_mask) ;

}
