#ifndef wncImageSource_H_
#define wncImageSource_H_

/*
 *
 * videoSpace-plugins/vnc/vncImageSource.H --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/image/source/ImageSource.H>
#include <nucleo/network/tcp/TcpConnection.H>
#include <nucleo/core/URI.H>
#include <nucleo/core/FileKeeper.H>

#include <X11/Xmd.h>
#include <X11/Xutil.h>

#ifdef HAVE_XCURSOR
#include <X11/Xcursor/Xcursor.h>
#else
typedef unsigned int XcursorPixel;
#endif

class XwncDesktop;

extern "C" {
#include <wncproto.h>
}

// ------------------------------------------------------------------
// See wncauth/wncproto.h

class wncSource : public ReactiveObject {

 private:

	std::string _hostname ;
	int _port ;
	std::string _password ;

	Image::Encoding _texEncoding ;
	Image::Encoding _wncEncoding ;
	int _depth;

	TcpConnection *_conn ;
	FileKeeper *_fk ;
	std::string _buffer ;

	XwncDesktop *_wncDesktop;

	unsigned int _width;
	unsigned int _height;

	bool _useShm;

	void _receive(char *data, unsigned int length);
	bool _ReadFromRFBServer(char *out, unsigned int n);
	bool _framebufferUpdate(rfbServerToClientMsg msg);
	bool _readWNCServer(void) ;

 protected:

	void react(Observable*);

 public:

	wncSource(
		XwncDesktop *wncDesktop, bool force_rgb_tex, const URI& uri,
		bool use_shm);
	~wncSource(void) { stop() ; }

 	bool start(void) ;
	void wncSource::startRequest(void);
	bool isActive(void) { return (_conn!=0) ; }
	bool stop(void) ;

	void updateRequest(bool include=true) ;
	void updateRequest(int x, int y, int w, int h, bool incremental=true) ;
	void keyEvent(unsigned long keysym, bool down_flag=true) ;
	void pointerEvent(
		Window id, int x, int y, unsigned char button_mask=0) ;
	void pointerEvent(int x, int y, unsigned char button_mask=0) ;
	void getSize(unsigned int *width, unsigned int *height);

} ;

#endif
