#ifndef AbstractWindowRenderer_H
#define AbstractWindowRenderer_H

/*
 *
 * wstyle/AbstractWindowRenderer.H --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/gl/scenegraph/sgNode.H>
#include <nucleo/core/ReactiveObject.H>

#include <X11/Xmd.h>
#include <X11/Xutil.h>

//#include "FoldableWindow.H"

using namespace nucleo ;

class XwncWindow ;

class AbstractWindowRenderer : public sgNode, public ReactiveObject {

protected:

  XwncWindow  *_window ;
  bool _dimmed ;
  bool _shadow ;

  int _nShapeRects;
  CARD16 *_shape;
  bool _canUseStencilBuffer;
  
  //void pseudoSelect(void) { select(); };

public:

  AbstractWindowRenderer(XwncWindow  *window, std::string name) ;
  ~AbstractWindowRenderer(void) {
    if (_shape) free(_shape);
  }
  XwncWindow *getWindow(void) { return _window ; }

  static bool dimmedByDefault ;
  static bool shadowByDefault ;

  void dim(void) {
    _dimmed = true ;
    _changed = true ;
  }
  void undim(void) {
    _dimmed = false ;
    _changed = true ;
  }

  void shadow(void) {
    _shadow = !_shadow ;
    _changed = true ;
  }

  void setShape(CARD16 *buf, int nrecs, bool cusb);
  bool hasShape(void);

  bool isInsideWindow(int x, int y);
  bool isInsideWindow2(int x, int y);

  virtual void keyEvent(unsigned long key, bool down_flag=true) ;
  virtual void pointerEvent(int x, int y, unsigned char button_mask=0) ;

  // for "foldable" window
  virtual void cropToCircule(void) {};
  virtual void testPolygon(void) {};
  virtual void simpleReset(void) {};
  virtual void fold(int px, int py, bool fold) {};
  virtual void animate(void) {};
} ;

AbstractWindowRenderer *createWindowRenderer(std::string style, XwncWindow *window) ;

#endif
