/*
 *
 * wstyle/AbstractWindowRenderer.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "config.h"

#include "AbstractWindowRenderer.H"
#include "wncdesktop/XwncWindow.H"

#include "FoldableWindow.H"

#include <stdexcept>

// ----------------------------------------------------------------------

bool AbstractWindowRenderer::dimmedByDefault = true ;
bool AbstractWindowRenderer::shadowByDefault = true ;

AbstractWindowRenderer::AbstractWindowRenderer(
	XwncWindow *window, std::string name) : sgNode(name)
{
	_window = window ;
	_dimmed = dimmedByDefault ;
	_shadow = shadowByDefault ;
	_shape = NULL;
	_nShapeRects = 0;
	_canUseStencilBuffer = false;
}

void AbstractWindowRenderer::setShape(CARD16 *buf, int nrecs, bool cusb) {
  if (_shape != NULL) {
    free(_shape);
    _shape = NULL;
  }
  _shape = buf;
  _nShapeRects = nrecs;
  _canUseStencilBuffer = cusb;
  postRedisplay() ;
}

bool AbstractWindowRenderer::hasShape(void) {
  if (_nShapeRects)
    return true;
  return false;
}

bool AbstractWindowRenderer::isInsideWindow(int x, int y) {

	float width,height,rx,ry;
	_window->getSize(&width, &height);
	_window->getRealPosition(&rx, &ry);

	float realx = x + width/2.0 + rx;
	float realy = -y + height/2.0 + ry;

	if (realx < rx || realx > rx + width || realy < ry ||
	    realy > ry + height)
	{
		return false;
	}

	if (!_nShapeRects || !_shape)
		return true;

	int i;
	for (i = 0; i < _nShapeRects; i++)
	{
		float x = _shape[i*4] + rx;
		float y = _shape[i*4 + 1] + ry;
		float w = _shape[i*4 + 2];
		float h = _shape[i*4 + 3];

		if (realx >= x && realx <= x+w && realy >= y && realy <= y + h)
		{
			return true;
		}
	}
	return false;
}

bool AbstractWindowRenderer::isInsideWindow2(int x, int y) {

	float width,height,rx,ry;
	_window->getSize(&width, &height);
     
	if (x < 0 || x > width || y < 0 || y > height)
	{
		return false;
	}

	if (!_nShapeRects || !_shape)
		return true;

	int i;
	for (i = 0; i < _nShapeRects; i++)
	{
		float xx = _shape[i*4];
		float yy = _shape[i*4 + 1];
		float w = _shape[i*4 + 2];
		float h = _shape[i*4 + 3];

		if (x >= xx && x <= xx+w && y >= yy && y <= yy + h)
		{
			return true;
		}
	}
	return false;
}

void
AbstractWindowRenderer::keyEvent(unsigned long key, bool down_flag)
{
	_window->keyEvent(key, down_flag) ;
}

void
AbstractWindowRenderer::pointerEvent(int x, int y, unsigned char button_mask)
{
	_window->pointerEvent(x,y,button_mask) ;
}

// ----------------------------------------------------------------------

AbstractWindowRenderer *
createWindowRenderer(std::string style, XwncWindow *window) {
  if (style=="foldable") return new FoldableWindow(window) ;

  std::string msg = std::string("Unknown window style: ")+style ;
  throw std::runtime_error(msg) ;
}
